/*
 *  This file is part of ixnet.library for the Amiga.
 *  Copyright (C) 2013 by Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: bpf.c,v 1.2 2015/01/19 20:38:02 piru Exp $
 *
 */

#define _KERNEL
#include "ixnet.h"

#include <machine/param.h>
#include <string.h>
#include <inetd.h>
#include <stdlib.h>
#include <amitcp/socketbasetags.h>
#include "select.h"
#include "ixprotos.h"

#if USE_BPF

#define f_handle f_so

int _bpf_read   (struct file *fp, char *buf, int len);
int _bpf_write  (struct file *fp, char *buf, int len);
int _bpf_ioctl  (struct file *fp, int cmd, int inout, int arglen, caddr_t data);
int _bpf_close  (struct file *fp);
int _bpf_select (struct file *fp, int select_cmd, int io_mode, fd_set *, u_long *);

static int
chk_berkeleypacketfilter(const char *name)
{
    if (strncmp(name, "/dev/bpf", 8) != 0)
	return -1;

    name += 8;

    if (!(name[0] >= '0' && name[0] <= '9' && (name[1] == 0 || (name[0] >= '1' && name[1] >= '0' && name[1] <= '9' && name[2] == 0))))
	return -1;

    if (name[1])
	return (name[0] - '0') * 10 + name[1] - '0';
    else
	return name[0] - '0';
}

int
_bpf_open (const char *name, int flags, int perms)
{
    usetup;
    struct ixnet *p = (struct ixnet *)u.u_ixnet;
    int chan;
    int err = -1;

    chan = chk_berkeleypacketfilter (name);
    if (chan > -1)
      {
	int network_protocol = p->u_networkprotocol;
	int32_t num_channels;

	switch (network_protocol)
	  {
	    case IX_NETWORK_AS225:
	    default:
		errno = ENODEV;
		break;
	    case IX_NETWORK_AMITCP:
		num_channels = 0;
		if (TCP_SocketBaseTags (SBTM_GETREF(SBTC_NUM_PACKET_FILTER_CHANNELS),(LONG)&num_channels,TAG_DONE) == 0 &&
		    num_channels > 0)
		  {
		    /* Instead of mapping specific device unit to specific channel,
		     * we rather just allocate the next available channel. - Piru
		     */
		    err = TCP_bpf_open (-1);
		    if (err != -1)
			TCP_bpf_set_interrupt_mask (err, SIGBREAKF_CTRL_C);
		  }
		  else
		    errno = ENODEV;
	        break;
	  }
      }
    else
	errno = ENOENT;
    return err;
}

int
_bpf_read (struct file *fp, char *buf, int len)
{
    usetup;
    struct user *usr = &u;
    struct ixnet *p = (struct ixnet *)usr->u_ixnet;
    int ostat, rc;

    ostat = usr->p_stat;
    usr->p_stat = SWAIT;

    rc = TCP_bpf_read (fp->f_handle,buf,len);

    if (CURSIG (usr))
	SetSignal (0, SIGBREAKF_CTRL_C);

    usr->p_stat = ostat;

    if (errno == EINTR)
	setrun (SysBase->ThisTask);

    return rc;
}

int
_bpf_write (struct file *fp, char *buf, int len)
{
    usetup;
    struct user *usr = &u;
    struct ixnet *p = (struct ixnet *)usr->u_ixnet;
    int ostat, rc;

    ostat = usr->p_stat;
    usr->p_stat = SWAIT;

    rc = TCP_bpf_write (fp->f_handle,buf,len);

    if (CURSIG (usr))
	SetSignal (0, SIGBREAKF_CTRL_C);

    usr->p_stat = ostat;

    if (errno == EINTR)
	setrun (SysBase->ThisTask);

    return rc;
}

int
_bpf_ioctl (struct file *fp, int cmd, int inout, int arglen, caddr_t data)
{
    usetup;
    struct user *usr = &u;
    struct ixnet *p = (struct ixnet *)usr->u_ixnet;
    int ostat, err = 0;

    ostat = usr->p_stat;
    usr->p_stat = SWAIT;

    err = TCP_bpf_ioctl (fp->f_handle,cmd,data);

    if (CURSIG (usr))
	SetSignal (0, SIGBREAKF_CTRL_C);

    usr->p_stat = ostat;

    if (errno == EINTR)
	setrun (SysBase->ThisTask);

    return err;
}

int
_bpf_close (struct file *fp)
{
    usetup;
    struct ixnet *p = (struct ixnet *)u.u_ixnet;

    return TCP_bpf_close (fp->f_handle);
}

int
_bpf_select (struct file *fp, int select_cmd, int io_mode, fd_set *set_notused, u_long *nfds_notused)
{
    usetup;
    struct ixnet *p = (struct ixnet *)u.u_ixnet;

    switch (select_cmd)
      {
	case SELCMD_PREPARE:
	    if (io_mode == SELMODE_IN)
	      {
		TCP_bpf_set_notify_mask (fp->f_handle, 1UL << p->u_sigio);
		return 1UL << p->u_sigio;
	      }
	    break;
	case SELCMD_CLEANUP:
	    if (io_mode == SELMODE_IN)
		TCP_bpf_set_notify_mask (fp->f_handle, 0);
	    break;
	case SELCMD_CHECK:
	case SELCMD_POLL:
	    switch (io_mode)
	      {
		case SELMODE_IN:
		    return TCP_bpf_data_waiting (fp->f_handle) == 0 ? 0 : 1;
		case SELMODE_OUT:
		    return 1; /* we're always ready for writing */
		case SELMODE_EXC:
		    return 0;
	      }
	    break;
      }
    return 0;
}

#else

int _bpf_open (const char *name, int flags, int perms) { usetup; errno = ENOSYS; return -1; }
int _bpf_read   (struct file *fp, char *buf, int len) { usetup; errno = ENOSYS; return -1; }
int _bpf_write  (struct file *fp, char *buf, int len) { usetup; errno = ENOSYS; return -1; }
int _bpf_ioctl  (struct file *fp, int cmd, int inout, int arglen, caddr_t data) { usetup; errno = ENOSYS; return -1; }
int _bpf_close  (struct file *fp) { usetup; errno = ENOSYS; return -1; }
int _bpf_select (struct file *fp, int select_cmd, int io_mode, fd_set *set_notused, u_long *nfds_notused) { usetup; errno = ENOSYS; return -1; }

#endif
