/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"

#include <string.h>
#include <stdlib.h>
#include <alloca.h>
#include <unistd.h>

/* __alloca.c -- allocate automatically reclaimed memory
   (Mostly) portable public-domain implementation -- D A Gwyn

   This implementation of the PWB library alloca function,
   which is used to allocate space off the run-time stack so
   that it is automatically reclaimed upon procedure exit,
   was inspired by discussions with J. Q. Johnson of Cornell.

   There are some preprocessor constants that can
   be defined when compiling for your specific system, for
   improved efficiency; however, the defaults should be okay.

   The general concept of this implementation is to keep
   track of all alloca-allocated blocks, and reclaim any
   that are found to be deeper in the stack than the current
   invocation.  This heuristic does not reclaim storage as
   soon as it becomes invalid, but it will do so eventually.

   As a special case, alloca(0) reclaims storage without
   allocating any.  It is a good idea to use alloca(0) in
   your main control loop, etc. to force garbage collection.  */

void
__alloca_cleanup(void)
{
	usetup;
	u_int sp = get_sp();
	struct alloca_mem *m, *prev = NULL;

	/* We assume here stack grow downwards. Remove all allocations which has
	 * lower depth than current sp */
	for(m = u.u_alloca; m != NULL; m = m->next)
	{
		if(m->depth < sp)
		{
			if(prev)
				prev->next = m->next;
			else
				u.u_alloca = m->next;

			free(m->addr);
			free(m);

			prev = m;
		}
		else
			break;
	}
}

void *
__alloca(size_t size)
{
	usetup;
	u_int sp = get_sp();

	/* Atempt to cleanup previous __alloca() resources */
	__alloca_cleanup();

	if((signed long)size > 0)
	{
		struct alloca_mem *a;
		void *res;
		a = malloc(sizeof(struct alloca_mem));
		res = valloc(size);
		if(a != NULL && res != NULL)
		{
			a->next  = u.u_alloca;
			a->addr  = res;
			a->depth = sp;
			u.u_alloca = a;
			return a->addr;
		}
		if(res)
			free(res);
		if(a)
			free(a);
	}

	return NULL;
}
