/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _KERNEL
#include <string.h>
#include "ixemul.h"

#if USE_BPF
extern int chk_berkeleypacketfilter(const char *name);
#endif

#if 0

struct lock_vec {
  int    mode;
  int    resolve_last;
};

static int
__lock_func (struct lockinfo *info, struct lock_vec *lv, int *error)
{
  struct StandardPacket *sp = &info->sp;

  sp->sp_Pkt.dp_Type = ACTION_LOCATE_OBJECT;
  sp->sp_Pkt.dp_Arg1 = info->parent_lock;
  sp->sp_Pkt.dp_Arg2 = info->bstr;
  sp->sp_Pkt.dp_Arg3 = lv->mode;

  PutPacket (info->handler, sp);
  __wait_sync_packet (sp);
  info->result = sp->sp_Pkt.dp_Res1;

  *error = info->result <= 0;

  /* continue if we failed because of symlink - reference ? */
  return lv->resolve_last;
}

BPTR __lock (const char *name, int mode)
{
  struct lock_vec lv;

  lv.mode = mode;
  lv.resolve_last = 1;

  return __plock (name, __lock_func, &lv);
}

BPTR __llock (const char *name, int mode)
{
  struct lock_vec lv;

  lv.mode = mode;
  lv.resolve_last = 0;

  return __plock (name, __lock_func, &lv);
}

#else

char *ix_to_ados(char *, const char *);
char *check_root(char *);

static int is_special_name(const char *name)
{
  if (!strcasecmp(name, "NIL:"))
    {
      SetIoErr(4242);
      return 1;
    }
  else if (is_pseudoterminal(name))
    {
      SetIoErr(5252);
      return 1;
    }
  else if (!strcmp(name, ":"))
    {
      SetIoErr(6262);
      return 1;
    }

  return 0;
}

BPTR __lock (const char *name, int mode)
{
  BPTR lock;
  APTR win;
  char *buf;
#if USE_BPF
  if (chk_berkeleypacketfilter(name) > -1)
  {
    SetIoErr(4242);
    return 0;
  }
#endif
  buf = alloca(strlen(name) + 4);
  buf = ix_to_ados(buf, name);
  if (is_special_name(buf))
    return 0;
  win = ((struct Process *)SysBase->ThisTask)->pr_WindowPtr;
  ((struct Process *)SysBase->ThisTask)->pr_WindowPtr = (APTR)-1;
  lock = Lock(buf, mode);
  if (lock == 0)
    {
      LONG err = IoErr();
      if (err == ERROR_OBJECT_NOT_FOUND)
        {
#warning "Currently only works for ../../../../ not something/../../../.. or ../../../../something"
	  buf = check_root(buf);
	  if (buf)
	    {
	      if (*buf == 0)
		err = 6262;
	      else
		{
		  lock = Lock(buf, mode);
		  if (lock)
		    err = 0;
		}
	    }

	  SetIoErr(err);
	}
    }
  ((struct Process *)SysBase->ThisTask)->pr_WindowPtr = win;
  return lock;
}

static BPTR __internal__llock (char *buf, int mode)
{
  char *filepos, oldchar, *buf2;
  BPTR dirlock, lock = 0;
  struct FileLock *fl;
  struct DosList *dl;
  LONG error;
  int len;
  APTR win;

  if (is_special_name(buf))
    {
      // IoErr has been set to special value here
      return 0;
    }

  filepos = FilePart(buf);
  len = strlen(filepos);
  if (len > 255)
    {
      SetIoErr(ERROR_LINE_TOO_LONG);
      return 0;
    }

  win = ((struct Process *)SysBase->ThisTask)->pr_WindowPtr;
  ((struct Process *)SysBase->ThisTask)->pr_WindowPtr = (APTR)-1;

#warning "Currenly doesn't handle multiassign - to implement would need to dup dos/Lock behavior"
#warning "Then again, neither did the original..."

  oldchar = *filepos;
  *filepos = '\0';
  dirlock = Lock(buf, ACCESS_READ);
  *filepos = oldchar;

  if (dirlock)
    {
#ifdef __MORPHOS__
      buf2 = alloca(len + 1); /* NOTE: MorphOS has stack always aligned at least by 4 */
#else
      buf2 =(char *) (((IPTR) alloca(len + 4) + 3) & (IPTR) -4);
#endif
      buf2[0] = len;
      memcpy(buf2 + 1, filepos, len);

      fl = BADDR(dirlock);
      dl = BADDR(fl->fl_Volume);

      lock = DoPkt3(dl->dol_Task, ACTION_LOCATE_OBJECT, dirlock, MKBADDR(buf2), mode);

      error = IoErr();
      UnLock(dirlock);
      SetIoErr(error);
    }

  ((struct Process *)SysBase->ThisTask)->pr_WindowPtr = win;
  return lock;
}

BPTR __llock (const char *name, int mode)
{
  char *buf = alloca(strlen(name) + 4);
  BPTR lock = 0;
  int last_err = 0;

#if USE_BPF
  if (chk_berkeleypacketfilter(name) > -1)
  {
    SetIoErr(4242);
    return 0;
  }
#endif

  buf = ix_to_ados(buf, name);

  lock = __internal__llock(buf, mode);

  if (lock == 0)
    {
      last_err = IoErr();
      if (last_err == ERROR_OBJECT_NOT_FOUND)
      {
#warning "Currently only works for ../../../../ not something/../../../.. or ../../../../something"
	buf = check_root(buf);
	if (buf)
	{
	  if (*buf == 0)
	  {
	    last_err = 6262;
	  }
	  else
	  {
	    lock = __internal__llock(buf, mode);
	  }
	}
      }
    }

  if (lock == 0)
    SetIoErr(last_err);

  return lock;
}

#endif
