/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *  Portions Copyright (C) 1996 by Jeff Shepherd
 *  Portions Copyright (C) 1996 by Hans Verkuil
 *  Portions Copyright (C) 2013 Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _KERNEL
#define DEBUG_VERSION

#include "ixemul.h"
#include "unp.h"
#include "kprintf.h"

#include <sys/ioctl.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <machine/param.h>
#include <string.h>

#if USE_BPF

#define f_handle f_so

static int poo_read   (struct file *fp, char *buf, int len);
static int poo_write  (struct file *fp, char *buf, int len);
static int poo_ioctl  (struct file *fp, int cmd, int inout, int arglen, caddr_t data);
static int poo_select (struct file *fp, int select_cmd, int io_mode, fd_set *ignored, u_long *also_ignored);
static int poo_close  (struct file *fp);

static void bpf_cleanup (int ostat)
{
    usetup;

    if (CURSIG (&u))
	SetSignal (0, SIGBREAKF_CTRL_C);

    u.p_stat = ostat;

    if (errno == EINTR)
	setrun (SysBase->ThisTask);

    KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
}

int
bpf_open (struct file *fp, const char *name, int mode, int perms)
{
    int ostat, omask;
    usetup;

    if (!u.u_ixnetbase)
    {
	errno = ENOSYS;
	KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	return -1;
    }

    ostat = u.p_stat;
    u.p_stat = SWAIT;
    omask = syscall (SYS_sigsetmask, ~0);

    do
      {
	fp->f_handle = netcall(NET__bpf_open, name, mode, perms);
	KPRINTF (("f_handle = %ld\n", fp->f_handle));

	if (fp->f_handle == -1)
	  {
	    break;
	  }

        /* set parameters */
        fp->f_stb.st_mode = 0666 | S_IFCHR;
        fp->f_stb.st_size = 1;
        fp->f_stb.st_blksize = 128;
        fp->f_flags = FREAD | FWRITE;
        fp->f_type = DTYPE_BPF;

        fp->f_read   = poo_read;
        fp->f_write  = poo_write;
        fp->f_ioctl  = poo_ioctl;
        fp->f_close  = poo_close;
        fp->f_select = poo_select;
      }
    while (0);

    syscall (SYS_sigsetmask, omask);
    bpf_cleanup(ostat);
    return fp->f_handle == -1 ? -1 : 0;
}

static int
poo_read (struct file *fp, char *buf, int len)
{
    usetup;

    return netcall(NET__bpf_read, fp, buf, len);
}

static int
poo_write (struct file *fp, char *buf, int len)
{
    usetup;

    return netcall(NET__bpf_write, fp, buf, len);
}

static int
poo_ioctl (struct file *fp, int cmd, int inout, int arglen, caddr_t data)
{
    usetup;

    return netcall(NET__bpf_ioctl, fp, cmd, inout, arglen, data);
}

static int
poo_close (struct file *fp)
{
    int err;
    usetup;

    ix_lock_base ();
    err = --fp->f_count;
    ix_unlock_base ();

    if (err)
	return 0;

    err = netcall(NET__bpf_close, fp);

    ffree(fp);

    KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
    return err;
}

static int
poo_select(struct file *fp, int select_cmd, int io_mode,
	   fd_set *ignored, u_long *also_ignored)
{
    usetup;

    return netcall(NET__bpf_select, fp, select_cmd, io_mode, ignored, also_ignored);
}

#endif
