/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 2015  Harry Sintonen.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define _KERNEL
#include <string.h>
#include <stdio.h>
#include "ixemul.h"
#include "kprintf.h"

extern const char SYS_XATTR_AMIGA_MODE[];
extern const char SYS_XATTR_AMIGA_COMMENT[];

extern int
__xattr_examine(const char *path, struct FileInfoBlock *fib, int followsoftlink);

static ssize_t
__xattr_fillattr(struct FileInfoBlock *fib, const char *name, void *value, size_t size)
{
  const int query = size == 0;
  size_t actual;
  usetup;

  if (!strcmp(name, SYS_XATTR_AMIGA_MODE))
    {
      actual = 4;
      if (query)
        return actual;
      else if (size >= actual)
        {
          uint8_t *data = value;
          data[0] = fib->fib_Protection >> 24;
          data[1] = fib->fib_Protection >> 16;
          data[2] = fib->fib_Protection >> 8;
          data[3] = fib->fib_Protection >> 0;
          return actual;
        }
      else
        {
          errno = ERANGE;
          return -1;
        }
    }
  else if (!strcmp(name, SYS_XATTR_AMIGA_COMMENT))
    {
      actual = strlen(fib->fib_Comment) + 1;
      if (query)
        return actual;
      else if (size >= actual)
        {
          memcpy(value, fib->fib_Comment, actual);
          return actual;
        }
      else
        {
          errno = ERANGE;
          return -1;
        }
    }

  errno = ENOATTR;
  return -1;
}

ssize_t
getxattr(const char *path, const char *name, void *value, size_t size)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 1) == 0)
    return __xattr_fillattr(&fib, name, value, size);
  return -1;
}

ssize_t
lgetxattr(const char *path, const char *name, void *value, size_t size)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 0) == 0)
    return __xattr_fillattr(&fib, name, value, size);
  return -1;
}

ssize_t
fgetxattr (int fd, const char *name, void *value, size_t size)
{
  struct file *fp;
  struct FileInfoBlock fib;
  ssize_t res = 0;
  int omask;
  BPTR lock;
  usetup;

  if (fd >= NOFILE || (fp = u.u_ofile[fd]) == NULL)
    {
      errno = EBADF;
      return -1;
    }

  /* no external attrs for null handler / sockets */
  if (HANDLER_NIL(fp) ||
      (fp->f_type == DTYPE_SOCKET && u.u_ixnetbase))
    {
      errno = ENOTSUP;
      return -1;
    }

  omask = syscall (SYS_sigsetmask, ~0);

  switch (fp->f_type)
    {
    case DTYPE_FILE:
      if (ExamineFH(CTOBPTR(fp->f_fh), &fib))
        res = __xattr_fillattr(&fib, name, value, size);
      else
        errno = __ioerr_to_errno (IoErr());
      break;
    case DTYPE_MEM:
      if (fp->f_name)
        {
          lock = __lock(fp->f_name, ACCESS_READ);
          if (lock)
            {
              if (Examine(lock, &fib))
                res = __xattr_fillattr(&fib, name, value, size);
             else
               errno = __ioerr_to_errno (IoErr());
            UnLock(lock);
          }
          else
            errno = __ioerr_to_errno (IoErr());
        }
      else
        errno = EINVAL;
      break;
    default:
      errno = ENOTSUP;
      break;
    }

  syscall (SYS_sigsetmask, omask);

  return res;
}
