/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"
#include <hardware/intbits.h>
#include "multiuser.h"
#include <string.h>

void
__ix_close_muFS( struct user *ix_u )
{
  if (muBase)
    {
      if (ix_u->u_UserInfo)
	{
	  muFreeUserInfo(ix_u->u_UserInfo);
	  ix_u->u_UserInfo = NULL;
	}

      if (ix_u->u_fileUserInfo)
	{
	  muFreeUserInfo(ix_u->u_fileUserInfo);
	  ix_u->u_fileUserInfo = NULL;
	}

      if (ix_u->u_GroupInfo)
	{
	  muFreeGroupInfo(ix_u->u_GroupInfo);
	 ix_u->u_GroupInfo = NULL;
	}

      if (ix_u->u_fileGroupInfo)
	{
	  muFreeGroupInfo(ix_u->u_fileGroupInfo);
	  ix_u->u_fileGroupInfo = NULL;
	}

      /* log me out */
      while (ix_u->u_setuid--)
	muLogout(muT_Quiet, TRUE, TAG_DONE);
    }
}

void __ix_set_stack_cookie(void *stacklow, size_t stacksize)
{
  uint32_t *ptr = stacklow;
  uint32_t i;
  if (stacksize >= sizeof(*ptr))
  {
    ptr++;
    stacksize -= sizeof(*ptr);
    for (i = 0; i < stacksize; i += sizeof(*ptr), ptr++)
    {
     *ptr = (uint32_t) ptr ^ 0xdbdbdbdb;
    }
  }
}

uint32_t __ix_get_stack_usage(const void *stacklow, size_t stacksize)
{
  const uint32_t *ptr = stacklow;
  uint32_t usage = stacksize;
  uint32_t i;
  if (stacksize >= sizeof(*ptr))
  {
    /* Skip the lowest 4 bytes of stack since it might contain a different stack cookie */
    ptr++;
    stacksize -= sizeof(*ptr);
    for (i = 0; i < stacksize; i += sizeof(*ptr), ptr++)
    {
      if (*ptr != ((uint32_t) ptr ^ 0xdbdbdbdb))
      {
	usage = stacksize - ((char *) --ptr - (char *) stacklow);
	break;
      }
    }
  }
  return usage;
}

void ix_stack_usage(void)
{
  /* Does someone want the stack usage data we have? */
  if (ix.ix_flags & ix_show_stack_usage)
    {
      struct Task *me = SysBase->ThisTask;
      struct user *u_ptr = getuser(me);
      BPTR lock = GetProgramDir();
      struct SUMessage sum;
      struct MsgPort *port, *reply;
      if ((reply = (struct MsgPort *)ix_create_port(0, 0)))
	{
	  bzero(&sum, sizeof(sum));
	  if (lock)
	    {
	      if (NameFromLock(lock, sum.name, sizeof(sum.name) - 40))
		strcat(sum.name, "/");
	    }
	  GetProgramName(sum.name + strlen(sum.name), 39);
	  NewGetTaskAttrsA(me, &sum.stack_size, sizeof(ULONG), TASKINFOTYPE_STACKSIZE_M68K, NULL);
	  sum.stack_usage = __ix_get_stack_usage(me->tc_SPLower, sum.stack_size);
	  if (u.u_68k_stk_max > sum.stack_size)
	    sum.stack_size = u.u_68k_stk_max;
          if (u.u_68k_stk_usage_highest > sum.stack_usage)
	    sum.stack_usage = u.u_68k_stk_usage_highest;
	  NewGetTaskAttrsA(me, &sum.ppc_stack_size, sizeof(ULONG), TASKINFOTYPE_STACKSIZE, NULL);
	  sum.ppc_stack_usage = __ix_get_stack_usage(me->tc_ETask->PPCSPLower, sum.ppc_stack_size);
	  if (u.u_ppc_stk_max > sum.ppc_stack_size)
	    sum.ppc_stack_size = u.u_ppc_stk_max;
          if (u.u_ppc_stk_usage_highest > sum.ppc_stack_usage)
	    sum.ppc_stack_usage = u.u_ppc_stk_usage_highest;
	  sum.msg.mn_Node.ln_Type = NT_MESSAGE;
	  sum.msg.mn_Length = sizeof(sum);
	  sum.msg.mn_ReplyPort = reply;
	  Forbid();
	  port = FindPort("ixstack port");
	  if (port)
	    PutMsg(port, &sum.msg);
	  Permit();
	  if (port)
	    WaitPort(reply);
	  ix_delete_port(reply);
	}
    }
}

void
ix_close (struct ixemul_base *ixbase)
{
  struct Task           *me = SysBase->ThisTask;
  struct user           *u_ptr = getuser(me);
  struct user           *ix_u = &u;
  struct Process        *child;
  struct user           *cu;
  struct ixnode         *dm;    /* really struct death_msg * */
  struct Message	*wb_msg;
  pid_t                 mypid;

  KPRINTF(("closing ixemul...\n"));

#ifdef NATIVE_MORPHOS
  NewGetTaskAttrsA(NULL, &mypid, sizeof(mypid), TASKINFOTYPE_PID, NULL);
  /* If running modern exec call the TLS destructors (if any) before doing
     anything else. This ensures that calling free() etc will still work
     right. - Piru */
  if (LIB_MINVER(&SysBase->LibNode, 51, 46))
    {
      TLSCallDestructors(NULL);
    }
#else
  mypid = (pid_t) me;
#endif

  Disable();
  ixremove(&timer_task_list, &ix_u->u_user_node);
  Enable();

#ifndef NATIVE_MORPHOS
  RemIntServer (INTB_VERTB, &ix_u->u_itimerint);
#endif

  Forbid();
  semexit(me);
  Permit();

#ifndef NOTRAP
  /* already reset the trap vector here. It's better to get an alert than
   * to loop infinitely if one of the following functions should crash */
  me->tc_TrapCode = ix_u->u_otrap_code;
#endif

  freestack();
  shmexit(ix_u);

  /* had to move this block after the SYS_close's, since close() might have
     to wait for a packet, and then it's essential that our switch/launch
     handlers are still active */
  SetExcept(0,~0);
  me->tc_ExceptCode = ix_u->u_oexcept_code;
  SetExcept(ix_u->u_oexcept_sigs, ~0);

  /* Make sure there is no exception pending. This can only happen if we
     are in Forbid or Disable state. */
  if (SysBase->TDNestCnt >= 0 || SysBase->IDNestCnt >= 0)
      me->tc_Flags &= ~TF_EXCEPT;

  KPRINTF(("TDNestCnt = %ld, IDNestCnt = %ld, SigExcept = %08lx, Flags = %lx\n",
	   SysBase->TDNestCnt, SysBase->IDNestCnt, me->tc_SigExcept, me->tc_Flags));

  FreeSignal (ix_u->u_sleep_sig);
  FreeSignal (ix_u->u_pipe_sig);

  if (ix_u->u_ixnetbase)
    {
      KPRINTF(("closing ixnet\n"));
      CloseLibrary (ix_u->u_ixnetbase);
    }

  KPRINTF(("closing timer.device\n"));
  CloseDevice ((struct IORequest *) ix_u->u_time_req);
  KPRINTF(("delete timer req\n"));
  ix_delete_extio((struct IORequest *)ix_u->u_time_req);

  if (ix_u->u_startup_cd != (BPTR) -1)
    {
      KPRINTF(("UnLock current dir\n"));
      __unlock (CurrentDir (ix_u->u_startup_cd));
      KPRINTF(("set dir name\n"));
      set_dir_name_from_lock(ix_u->u_startup_cd);
    }

  ix_u->u_trace_flags = 1;
  KPRINTF(("delete select port\n"));
  ix_delete_port(ix_u->u_select_mp);
  KPRINTF(("delete sync port\n"));
  ix_delete_port(ix_u->u_sync_mp);

  /* try to free it here */
  KPRINTF(("closing mufs\n"));
  __ix_close_muFS(ix_u);

  KPRINTF(("relinking processes\n"));
  Forbid();
  for ((child = ix_u->p_cptr); child; (child = cu->p_osptr))
    {
KPRINTF(("child = %lx\n", child));
      cu = safe_getuser(child);
KPRINTF(("user = %lx\n", cu));
      cu->p_pptr = (struct Process *) 1;
    }

  ix_u->p_cptr = 0;

  /* decrease session count and possibly free the session structure */
  if (ix_u->u_session)
    {
      if (ix_u->u_session->pgrp == mypid)
	ix_u->u_session->pgrp = 0;
      if (ix_u->u_session->s_count-- <= 1)
	kfree(ix_u->u_session);
    }

KPRINTF(("ysptr = %lx\n", ix_u->p_ysptr));
  if (ix_u->p_ysptr)
    safe_getuser(ix_u->p_ysptr)->p_osptr = ix_u->p_osptr;

KPRINTF(("osptr = %lx\n", ix_u->p_osptr));
  if (ix_u->p_osptr)
    safe_getuser(ix_u->p_osptr)->p_ysptr = ix_u->p_ysptr;

KPRINTF(("pptr = %lx\n", ix_u->p_pptr));
  if (ix_u->p_pptr && ix_u->p_pptr != (struct Process *)1)
    {
      struct user *u_ptr = safe_getuser(ix_u->p_pptr);
KPRINTF(("user = %lx\n", u_ptr));
      if (u_ptr->p_cptr == (struct Process *)me)
	u_ptr->p_cptr = ix_u->p_osptr;
    }
  Permit();

  if (ix_u->p_flag & SFREEA4)
    {
      KPRINTF(("Freeing data\n"));
      kfree (ix_u->u_sdata_ptr);
    }

  KPRINTF(("waking up zombies\n"));
  for (dm = (struct ixnode *)ix_u->p_zombies.head; dm; )
    {
      struct ixnode *tmp = dm;
      dm = dm->next;
      /* there might be children sleeping on this, so wake them up now.. */
      ix_wakeup((u_int)tmp);
      kfree(tmp);
    }

  FreeSignal (ix_u->p_zombie_sig);

  if ((ix_u->p_flag & (SUSAGE_TRACKED | SUSAGE_DONE)) == SUSAGE_TRACKED)
    ix_stack_usage();

  KPRINTF(("Freeing all memory\n"));
  all_free ();

  /* delay this until here, since the above called functions need access
   * to the user area. */
#ifdef NOTRAP
  me->tc_UserData = ix_u->u_ouser;
#else
  me->tc_TrapData = ix_u->u_ouser;
#endif

  wb_msg = ix_u->u_wbmsg;
  kfree (((char *)ix_u) - ix_u->u_a4_pointers_size * 4);

  /* Mask out any pending CTRL-x signals. ATM CTRL-E is used by setrun.
   * CTRL-D should not be used by ixemul, and is reserved for aborting
   * AmigaDOS scripts. CTRL-C is cleared by shell. - Piru */
  SetSignal(0, SIGBREAKF_CTRL_E | SIGBREAKF_CTRL_F);

  if (wb_msg)
    {
      Forbid();
      ReplyMsg(wb_msg);
    }

  KPRINTF(("done.\n"));
}
