/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *  Copyright (C) 2020  Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  kern_descrip.c,v 1.1.1.1 1994/04/04 04:30:42 amiga Exp
 *
 *  kern_descrip.c,v
 * Revision 1.1.1.1  1994/04/04  04:30:42  amiga
 * Initial CVS check in.
 *
 *  Revision 1.2  1992/07/04  19:19:04  mwild
 *  add support for F_INTERNALIZE/F_EXTERNALIZE, which are used by IXPIPE
 *  and execve().
 *
 * Revision 1.1  1992/05/14  19:55:40  mwild
 * Initial revision
 *
 *
 *  Since the code originated from Berkeley, the following copyright
 *  header applies as well. The code has been changed, it's not the
 *  original Berkeley code!
 */

/*
 * Copyright (c) 1982, 1986, 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution is only permitted until one year after the first shipment
 * of 4.4BSD by the Regents.  Otherwise, redistribution and use in source and
 * binary forms are permitted provided that: (1) source distributions retain
 * this entire copyright notice and comment, and (2) distributions including
 * binaries display the following acknowledgement:  This product includes
 * software developed by the University of California, Berkeley and its
 * contributors'' in the documentation or other materials provided with the
 * distribution and in all advertising materials mentioning features or use
 * of this software.  Neither the name of the University nor the names of
 * its contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *      @(#)kern_descrip.c      7.16 (Berkeley) 6/28/90
 */

#define _KERNEL
#define EXEC_SILENT_BUILD
#include "ixemul.h"
#include "kprintf.h"
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <string.h>
#include <exec/avl.h>
#include <clib/savl_protos.h>

#undef tv_sec

/*
 * Descriptor management.
 */

static int
falloc_generic (struct file **resultfp, int *resultfd, int startfd);

/*
 * System calls on descriptors.
 */
/* ARGSUSED */
int
getdtablesize (void)
{
  /* This was NOFILE, but getdtablesize is also used to determine the number
     number of filehandles select() should test. And if you pass a value
     larger than FD_SETSIZE to select(), you'll get fireworks. And it is very
     hard to discover why your program won't work. I know: I've been through
     this process twice now, so I thought I'd better fix this here by just
     returning FD_SETSIZE. 
     
     I could also change the FD_SETSIZE macro to the value of NOFILE (changing
     256 to 512), but that would be a waste of memory, and besides, in the
     future the number of filehandles might become dynamic, so it wouldn't work
     in any case. */
  return FD_SETSIZE;
}

/*
 * Duplicate a file descriptor.
 */
static int
dup_generic (unsigned i, int startfd)
{
  struct file *fp;
  int fd, error;
  usetup;

  if (i >= NOFILE || (fp = u.u_ofile[i]) == NULL)
    errno_return(EBADF, -1);

  if (fp->f_type == DTYPE_SOCKET)
    {
      struct file *fp2;
      int fd2;
      int err;

      if ((err = falloc_generic (&fp2, &fd2, startfd)))
	errno_return(err, -1);

      fp2->f_so = netcall(NET__dup, fp);
      if (fp2->f_so == -1)
      {
	/* free the allocated fd */
	u.u_ofile[fd2] = 0;
	fp2->f_count = 0;
	ffree(fp2);
	return -1;
      }
      fp2->f_socket_domain = fp->f_socket_domain;
      fp2->f_socket_type = fp->f_socket_type;
      fp2->f_socket_protocol = fp->f_socket_protocol;
      _set_socket_params(fp2, fp->f_socket_domain, fp->f_socket_type, fp->f_socket_protocol);
      return fd2;
    }

  if ((error = ufalloc (startfd, &fd)))
    errno_return(error, -1);

  ix_lock_base ();

  u.u_ofile[fd] = fp;
  u.u_pofile[fd] = u.u_pofile[i] &~ UF_EXCLOSE;
  fp->f_count++;
  if (fd > u.u_lastfile)
    u.u_lastfile = fd;

  ix_unlock_base ();

  return fd;
}

/* ARGSUSED */
int
dup (unsigned i)
{
  return dup_generic (i, 0);
}

/*
 * Duplicate a file descriptor to a particular value.
 */
/* ARGSUSED */
int
dup2 (unsigned i, unsigned j)
{
  register struct file *fp;
  int old_err;
  usetup;

  if (i >= NOFILE || (fp = u.u_ofile[i]) == NULL)
    errno_return(EBADF, -1);

  if (j < 0 || j >= NOFILE)
    errno_return(EBADF, -1);

  if (i == j) return (int)j;

  ix_lock_base ();

  old_err = errno;
  if (u.u_ofile[j] && u.u_ofile[j]->f_close)
    u.u_ofile[j]->f_close (u.u_ofile[j]);

  u.u_ofile[j] = fp;
  u.u_pofile[j] = u.u_pofile[i] &~ UF_EXCLOSE;
  fp->f_count++;
  if (j > u.u_lastfile)
    u.u_lastfile = j;

  ix_unlock_base ();

  /*
   * dup2() must suceed even though the close had an error.
   */
  errno = old_err;      /* XXX */
  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
  return (int)j;
}

/*
 * The file control system call.
 */
/* ARGSUSED */
int
fcntl(int fdes, int cmd, int arg)
{
  register struct file *fp;
  register char *pop;
  int i, error;
  usetup;

  /* F_INTERNALIZE doesn't need a valid descriptor. Check for this first */
  if (cmd == F_INTERNALIZE)
    {
      if ((error = ufalloc (0, &i)))
	{
	  errno = error;
	  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	  return -1;
	}
      fp = (struct file *) arg;
      u.u_ofile[i] = fp;
      u.u_pofile[i] = 0;
      ix_lock_base ();
      fp->f_count++;
      ix_unlock_base ();
      return i;
    }


  if (fdes >= NOFILE || (fp = u.u_ofile[fdes]) == NULL)
    {
      errno = EBADF;
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      return -1;
    }

  pop = &u.u_pofile[fdes];
  switch (cmd) 
    {
      case F_DUPFD:
	return dup_generic(fdes, arg);

      case F_GETFD:
	return (*pop & UF_EXCLOSE) ? FD_CLOEXEC : 0;

      case F_SETFD:
	*pop = (*pop &~ UF_EXCLOSE) | ((arg & FD_CLOEXEC) ? UF_EXCLOSE : 0);
	return 0;

      case F_GETFL:
	return OFLAGS(fp->f_flags);

      case F_SETFL:
	fp->f_flags &= ~FCNTLFLAGS;
	fp->f_flags |= FFLAGS(arg) & FCNTLFLAGS;
	if (fp->f_type == DTYPE_SOCKET)
	{
	  arg = (fp->f_flags & FASYNC) ? 1 : 0;
	  ioctl(fdes, FIOASYNC, &arg);
	  arg = (fp->f_flags & FNONBLOCK) ? 1 : 0;
	  ioctl(fdes, FIONBIO, &arg);
	}
	return 0;

      case F_EXTERNALIZE:
	return (int)fp;

      case F_SETLK:
      case F_SETLKW:
      case F_GETLK:
        /* We don't support this broken POSIX lock mess. */
        errno = ENOLCK;
        return -1;

      default:
	errno = EINVAL;
	KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	return -1;
    }
	/* NOTREACHED */
}

/*
 * Return status information about a file descriptor.
 */

static inline void conv_stat64_to_stat(const struct stat64 *fsb, struct stat *tsb)
{
  bzero(tsb, sizeof(*tsb));

  tsb->st_dev     = fsb->st_dev;
  tsb->st_ino     = fsb->st_ino;
  tsb->st_mode    = fsb->st_mode;
  tsb->st_nlink   = fsb->st_nlink;
  tsb->st_uid     = fsb->st_uid;
  tsb->st_gid     = fsb->st_gid;
  tsb->st_rdev    = fsb->st_rdev;
  tsb->st_size    = fsb->st_size > 0x7fffffff ? 0 : fsb->st_size;
  tsb->st_atime   = fsb->st_atime;
  tsb->st_mtime   = fsb->st_mtime;
  tsb->st_ctime   = fsb->st_ctime;
  tsb->st_blksize = fsb->st_blksize;
  tsb->st_blocks  = fsb->st_blocks > 0x7fffffff ? 0 : fsb->st_blocks;
  /* amiga fields */
  tsb->st_handler = fsb->st_handler;
  tsb->st_amode   = fsb->st_amode;
}

/* ARGSUSED */
int
fstat (int fdes, struct stat *sb)
{
  struct file *fp;
  usetup;

  if (fdes >= NOFILE || (fp = u.u_ofile[fdes]) == NULL)
    {
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      errno = EBADF;
      return -1;
    }

  /* I found this code from the AmiTCP sdk. It *should* work with AS225 */
  if (fp->f_type == DTYPE_SOCKET && u.u_ixnetbase)
    return netcall(NET__fstat, fp, sb);

  /* stdin, stdout and stderr as pipe aren't really opened so cook up something */
  if (fdes >= 0 && fdes <= 2 && fp->f_type == DTYPE_PIPE)
    {
      bzero(&fp->f_stb, sizeof(&fp->f_stb));
      fp->f_stb.st_ino = INO_PIPE;
      fp->f_stb.st_mode = 0666 | S_IFCHR;
      fp->f_stb.st_size = UNIX_SOCKET_SIZE;
      fp->f_stb.st_blksize = 512;
    }
  /* if this file is writable, then it's quite probably that the stat buffer
     information stored at open time is no longer valid. So if the file really
     is a file, update that information */
  else if ((fp->f_flags & FWRITE) && fp->f_type == DTYPE_FILE)
    __fstat64 (fp);

  conv_stat64_to_stat(&fp->f_stb, sb);
  return 0;
}

/*
 * Return status information about a file descriptor.
 */
/* ARGSUSED */
int
fstat64 (int fdes, struct stat64 *sb)
{
  struct file *fp;
  usetup;

  if (fdes >= NOFILE || (fp = u.u_ofile[fdes]) == NULL)
    {
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      errno = EBADF;
      return -1;
    }

  /* I found this code from the AmiTCP sdk. It *should* work with AS225 */
  if (fp->f_type == DTYPE_SOCKET && u.u_ixnetbase)
    return netcall(NET__fstat64, fp, sb);

  /* if this file is writable, then it's quite probably that the stat buffer
     information stored at open time is no longer valid. So if the file really
     is a file, update that information */
  if ((fp->f_flags & FWRITE) && fp->f_type == DTYPE_FILE)
    __fstat64 (fp);

  *sb = fp->f_stb;
  return 0;
}

/*
 * Allocate a user file descriptor.
 */
int
ufalloc(int want, int *result)
{
  usetup;

  if (want < 0 || want >= NOFILE)
    return EINVAL;

  for (; want < NOFILE; want++) 
    {
      if (u.u_ofile[want] == NULL) 
	{
	  u.u_pofile[want] = 0;
	  if (want > u.u_lastfile) u.u_lastfile = want;
			
	  *result = want;
	  return 0;
	}
    }
  return EMFILE;
}

/*
 * Allocate a user file descriptor
 * and a file structure.
 * Initialize the descriptor
 * to point at the file structure.
 */
static int
falloc_generic (struct file **resultfp, int *resultfd, int startfd)
{
  register struct file *fp;
  int error, i;
  usetup;

  if ((error = ufalloc(startfd, &i)))
    return (error);

  ix_lock_base ();

#if USE_DYNFILETAB

  {
    struct MinNode *node;
    node = REMHEAD(&ix.ix_free_file_list);
    if (node)
    {
      fp = (void *) (node + 1);
      ADDHEAD(&ix.ix_used_file_list, node);
      goto slot;
    }

    node = kmalloc(sizeof(struct MinNode) + sizeof(struct file));
    if (node)
    {
      fp = (void *) (node + 1);
      memset(fp, 0, sizeof(struct file));
      ADDHEAD(&ix.ix_used_file_list, node);
      goto slot;
    }

    //ix_warning("ixemul.library out of memory!");
    error = ENOMEM;
  }

#else

  if (ix.ix_lastf == 0)
    ix.ix_lastf = ix.ix_file_tab;

  for (fp = ix.ix_lastf; fp < ix.ix_fileNFILE; fp++)
    if (fp->f_count == 0)
      goto slot;

  for (fp = ix.ix_file_tab; fp < ix.ix_lastf; fp++)
    if (fp->f_count == 0)
      goto slot;

  /* YES I know.. it's not optimal, we should resize the table...
   * unfortunately all code accessing file structures will then have
   * to be changed as well, and this is a job for later improvement,
   * first goal is to get this baby working... */

  ix_warning("ixemul.library file table full!");
  error = ENFILE;

#endif

  goto do_ret;

slot:
  u.u_ofile[i] = fp;
  fp->f_stb_dirty = 0;
  fp->f_name = 0;
  fp->f_count = 1;
  fp->f_type = 0;               /* inexistant type ;-) */
  memset(&fp->f__fh, 0, sizeof(fp->f__fh));
  fp->f_parentlock = 0;
  fp->f_flockkey.handler = 0;
#if !USE_DYNFILETAB
  ix.ix_lastf = fp + 1;
#endif
  if (resultfp)
    *resultfp = fp;
  if (resultfd)
    *resultfd = i;

  error = 0;

do_ret:
  ix_unlock_base();

  return error;
}

int
falloc(struct file **resultfp, int *resultfd)
{
  return falloc_generic (resultfp, resultfd, 0);
}


void ffree(struct file *f)
{
#if USE_DYNFILETAB
  if (f)
  {
    struct MinNode *node = ((struct MinNode *) f) - 1;

    ix_lock_base ();

    /* Remove from ix_used_file_list */
    REMOVE(node);

    ADDHEAD(&ix.ix_free_file_list, node);

    ix_unlock_base ();
  }
#endif
}


/*
 * Apply an advisory lock on a file descriptor.
 */
#define USE_NON_ATOMIC_RELOCK 0
//#define DFLOCK(...) dprintf(__VA_ARGS__)
#define DFLOCK(...) do {} while(0)

static struct AVLNode *_g_flocktree;

struct flocknode
{
  struct AVLNode  node;
  struct SignalSemaphore sema;
  struct AVLNode *tasktree;
  struct ix_flockkey key;
};

LONG flocknodecomp(const struct AVLNode *n1, const struct AVLNode *n2)
{
  const struct flocknode *node1 = (const struct flocknode *) n1;
  const struct flocknode *node2 = (const struct flocknode *) n2;
  if (node1->key.handler < node2->key.handler)
    return -1;
  else if (node1->key.handler > node2->key.handler)
    return 1;
  else if (node1->key.ino < node2->key.ino)
    return -1;
  else if (node1->key.ino > node2->key.ino)
    return 1;
  return 0;
}

LONG flockkeycomp(const struct AVLNode *n, AVLKey k)
{
  const struct flocknode *node = (const struct flocknode *) n;
  const struct ix_flockkey *key = (const struct ix_flockkey *) k;
  if (node->key.handler < key->handler)
    return -1;
  else if (node->key.handler > key->handler)
    return 1;
  else if (node->key.ino < key->ino)
    return -1;
  else if (node->key.ino > key->ino)
    return 1;
  return 0;
}

struct tasknode
{
  struct AVLNode node;
  ULONG          pid;
};

LONG tasknodecomp(const struct AVLNode *n1, const struct AVLNode *n2)
{
  const struct tasknode *node1 = (const struct tasknode *) n1;
  const struct tasknode *node2 = (const struct tasknode *) n2;
  if (node1->pid < node2->pid)
    return -1;
  else if (node1->pid > node2->pid)
    return 1;
  return 0;
}

LONG taskkeycomp(const struct AVLNode *n, AVLKey k)
{
  const struct tasknode *node = (const struct tasknode *) n;
  const ULONG key = (ULONG) k;
  if (node->pid < key)
    return -1;
  else if (node->pid > key)
    return 1;
  return 0;
}

int i_funlock(struct file *fp)
{
  ULONG pid;
  struct flocknode *flocknode;
  struct tasknode *tasknode;

  NewGetTaskAttrsA(NULL, &pid, sizeof(pid), TASKINFOTYPE_PID, TAG_DONE);

  /* Check if we hold the lock */
  if (fp->f_flockkey.handler == 0 ||
      (flocknode = (struct flocknode *) AVL_FindNode(_g_flocktree, (AVLKey) &fp->f_flockkey, flockkeycomp)) == NULL ||
      (tasknode = (struct tasknode *) AVL_RemNodeByKey(&flocknode->tasktree, (AVLKey) pid, taskkeycomp)) == NULL)
    {
      return -1;
    }

  DFLOCK("%s: task %p flocknode %p free tasknode %p\n", __func__, FindTask(NULL), flocknode, tasknode);
  kfree(tasknode);

  /* All locks gone? */
  if (flocknode->tasktree == NULL)
    {
      DFLOCK("%s: task %p free flocknode %p\n", __func__, FindTask(NULL), flocknode);
      AVL_RemNodeByAddress(&_g_flocktree, &flocknode->node);
      kfree(flocknode);
    }
  else
    ReleaseSemaphore(&flocknode->sema);

  /* Unlocked ok! */
  return 0;
}

/* ARGSUSED */
int
flock (int fdes, int how)
{
  usetup;
  struct user *p = &u;
  struct file *fp;
  struct Process *self;
  ULONG pid;
  int mode;
  struct stat st;
  struct flocknode *flocknode;
  struct tasknode *tasknode;
  struct SemaphoreMessage sm;
  struct MsgPort *port;
  ULONG recv_wait_sigs;
  int ostat;
  LONG rc;
  int pendingsig;

  mode = how & (LOCK_SH | LOCK_EX | LOCK_UN);
  if (mode != LOCK_SH && mode != LOCK_EX && mode != LOCK_UN)
    {
      errno = EINVAL;
      return -1;
    }

  if (fdes >= NOFILE || (fp = u.u_ofile[fdes]) == NULL)
    {
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      errno = EBADF;
      return -1;
    }

  if (fp->f_type != DTYPE_FILE ||
      fp->f_parentlock == 0 ||
      fp->f_name == NULL)
    {
      errno = EOPNOTSUPP;
      return -1;
    }

  self = (struct Process *) FindTask(NULL);

  ix_lock_base ();

  if (how & LOCK_UN)
    {
      rc = i_funlock(fp);
      ix_unlock_base ();
      if (rc == -1)
        errno = EINVAL;
      return rc;
    }

  if (fp->f_flockkey.handler == 0)
    {
      rc = syscall(SYS_fstat, fdes, &st) == 0 && S_ISREG(st.st_mode) && st.st_handler;
      if (!rc)
        {
          ix_unlock_base ();
          errno = EINVAL;
          return -1;
        }
      fp->f_flockkey.handler = st.st_handler;
      fp->f_flockkey.ino     = st.st_ino;
    }

  DFLOCK("%s: task %p flockkey %08x%08x\n", __func__, self, fp->f_flockkey.handler, fp->f_flockkey.ino);
  flocknode = (struct flocknode *) AVL_FindNode(_g_flocktree, (AVLKey) &fp->f_flockkey, flockkeycomp);
  DFLOCK("%s: task %p found flocknode #1 %p\n", __func__, self, flocknode);

  NewGetTaskAttrsA(NULL, &pid, sizeof(pid), TASKINFOTYPE_PID, TAG_DONE);

  if (flocknode)
    {
      tasknode = (struct tasknode *) AVL_FindNode(flocknode->tasktree, (AVLKey) pid, taskkeycomp);
      DFLOCK("%s: task %p flocknode %p found tasknode %p\n", __func__, self, flocknode, tasknode);
      if (tasknode)
        {
          if (((mode & LOCK_EX) && flocknode->sema.ss_Owner != NULL) ||
              ((mode & LOCK_SH) && flocknode->sema.ss_Owner == NULL))
            {
              /* Already in correct mode, do nothing */
              DFLOCK("%s: task %p already in correct mode, return success\n", __func__, self);
              ix_unlock_base ();
              return 0;
            }

          /* We hold a lock currently, unlock it */
          i_funlock(fp);
#if USE_NON_ATOMIC_RELOCK
          /* Make it possible for other things to lock while upgrading/downgrading */
          ix_unlock_base ();
          ix_lock_base ();
#endif
        }
    }

  flocknode = (struct flocknode *) AVL_FindNode(_g_flocktree, (AVLKey) &fp->f_flockkey, flockkeycomp);
  DFLOCK("%s: task %p found flocknode #2 %p\n", __func__, self, flocknode);

  if (!flocknode)
    {
      flocknode = kmalloc(sizeof(*flocknode));
      if (!flocknode)
        {
          ix_unlock_base ();
          errno = ENOLCK;
          return -1;
        }
      flocknode->tasktree = NULL;
      flocknode->key = fp->f_flockkey;
      memset(&flocknode->sema, 0, sizeof(&flocknode->sema));
      InitSemaphore(&flocknode->sema);
      AVL_AddNode(&_g_flocktree, &flocknode->node, flocknodecomp);
    }

  /* Attempt to lock right away in non-blocking way */
  if (mode & LOCK_EX)
    rc = AttemptSemaphore(&flocknode->sema);
  else
    rc = AttemptSemaphoreShared(&flocknode->sema);

  if (!rc && (how & LOCK_NB))
    {
      /* Someone else is holding a lock, we would block */
      ix_unlock_base ();
      errno = EWOULDBLOCK;
      return -1;
    }

  tasknode = kmalloc(sizeof(*tasknode));
  if (!tasknode)
    {
      if (rc)
        ReleaseSemaphore(&flocknode->sema);
      /* All locks gone? */
      if (flocknode->tasktree == NULL)
        {
          AVL_RemNodeByAddress(&_g_flocktree, &flocknode->node);
          kfree(flocknode);
        }
      ix_unlock_base ();
      errno = ENOLCK;
      return -1;
    }

  tasknode->pid = pid;
  AVL_AddNode(&flocknode->tasktree, &tasknode->node, tasknodecomp);
  DFLOCK("%s: task %p flocknode %p added tasknode %p\n", __func__, self, flocknode, tasknode);

  /* If we got the lock already, we're done */
  if (rc)
    {
      DFLOCK("%s: task %p locked ok, return success\n", __func__, self);
      ix_unlock_base ();
      return 0;
    }

  /* If not, we must wait for the lock to become available or for a signal */
  port = &self->pr_MsgPort;
  sm.ssm_Message.mn_Node.ln_Type = NT_MESSAGE;
  sm.ssm_Message.mn_Node.ln_Name = (STRPTR) (mode & LOCK_EX ? 0 : 1);
  sm.ssm_Message.mn_ReplyPort = port;
  sm.ssm_Message.mn_Length = sizeof(sm);

  ostat = p->p_stat;
  p->p_stat = SSLEEP;
  p->p_wchan = (caddr_t) flock;
  p->p_wmesg = "flock";

  ix_unlock_base (); /* unlock here to make sure we can't be interrupted by a signal */

  Procure(&flocknode->sema, &sm);
  recv_wait_sigs = Wait((1UL << p->u_sleep_sig) | (1UL << port->mp_SigBit) | SIGBREAKF_CTRL_C);

  if (!(recv_wait_sigs & (1UL << port->mp_SigBit)))
    Vacate(&flocknode->sema, &sm);
  /* Clean up the port */
  WaitPort(port);
  GetMsg(port);

  pendingsig = issig(p);
  DFLOCK("%s: task %p pendingsig %d\n", __func__, self, pendingsig);

  if (pendingsig || (recv_wait_sigs & SIGBREAKF_CTRL_C))
    {
      ix_lock_base ();
      DFLOCK("%s: task %p signal %d! flocknode %p remove tasknode %p\n", __func__, self, pendingsig, flocknode, tasknode);
      if (sm.ssm_Semaphore)
        ReleaseSemaphore(&flocknode->sema);
      AVL_RemNodeByAddress(&flocknode->tasktree, &tasknode->node);
      kfree(tasknode);
      /* All locks gone? */
      if (flocknode->tasktree == NULL)
        {
          DFLOCK("%s: task %p signal! remove flocknode %p\n", __func__, self, flocknode);
          AVL_RemNodeByAddress(&_g_flocktree, &flocknode->node);
          kfree(flocknode);
        }
      ix_unlock_base ();
    }

  p->p_wchan = 0;
  p->p_wmesg = 0;
  p->p_stat = ostat;

  /* need special processing for ^C here, as that is completely disabled
     when we're SSLEEPing */
  if (recv_wait_sigs & SIGBREAKF_CTRL_C)
    _psignal(SysBase->ThisTask, SIGINT);
  setrun(SysBase->ThisTask);

  if (pendingsig || (recv_wait_sigs & SIGBREAKF_CTRL_C))
    {
      errno = EINTR;
      return -1;
    }

  /* All good */
  return 0;
}

