/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 2015  Harry Sintonen.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define _KERNEL
#include <string.h>
#include <stdio.h>
#include <libgen.h>
#include <stdlib.h>
#include "ixemul.h"
#include "kprintf.h"

const char SYS_XATTR_AMIGA_MODE[]    = "user.amiga.mode";
const char SYS_XATTR_AMIGA_COMMENT[] = "user.amiga.comment";

int
__xattr_examine(const char *path, struct FileInfoBlock *fib, int followsoftlink)
{
  int res = -1;
  BPTR (*funcptr)(char *name, int mode);
  int omask;
  BPTR lock;
  usetup;

  funcptr = followsoftlink ? __lock : __llock;

  omask = syscall (SYS_sigsetmask, ~0);

  lock = funcptr((STRPTR) path, ACCESS_READ);
  if (lock)
    res = Examine(lock, fib) == 0 ? -1 : 0;
  else if (!followsoftlink && IoErr() == ERROR_IS_SOFT_LINK)
    {
      /* There was a softlink somewhere within the path - but we
       * don't know where - split the path in dir and base names and
       * try again. If the base element is a softlink __llock will
       * fail with ERROR_IS_SOFT_LINK. Since we cannot easily examine
       * the link itself, this is the best behavior. */
      char *dirnamebuf = strdup(path);
      char *basenamebuf = strdup(path);
      if (dirnamebuf && basenamebuf)
        {
          BPTR dirlock = __lock(dirname(dirnamebuf), ACCESS_READ);
          if (dirlock)
            {
              BPTR olddirlock = CurrentDir(dirlock);
              lock = __llock(basename(basenamebuf), ACCESS_READ);
              if (lock)
                res = Examine(lock, fib) == 0 ? -1 : 0;

              CurrentDir(olddirlock);
              UnLock(dirlock);
            }
        }
      else
        SetIoErr(ERROR_NO_FREE_STORE);

      free(basenamebuf);
      free(dirnamebuf);
    }

  if (res == -1)
    errno = __ioerr_to_errno (IoErr());

  UnLock(lock);

  syscall (SYS_sigsetmask, omask);

  return res;
}

static size_t
__xattr_filllist(struct FileInfoBlock *fib, char *list, size_t size)
{
  size_t res = 0;
  const int query = size == 0;
  size_t actual;
  usetup;

  actual = sizeof(SYS_XATTR_AMIGA_MODE);
  if (query)
    ;
  else if (size >= actual)
    {
      memcpy(list, SYS_XATTR_AMIGA_MODE, actual);
      list += actual;
      size -= actual;
    }
  else
    {
      errno = ERANGE;
      return -1;
    }
  res += actual;

  if (fib->fib_Comment[0])
    {
      actual = sizeof(SYS_XATTR_AMIGA_COMMENT);
      if (query)
        ;
      else if (size >= actual)
        {
          memcpy(list, SYS_XATTR_AMIGA_COMMENT, actual);
          list += actual;
          size -= actual;
        }
      else
        {
          errno = ERANGE;
          return -1;
        }
      res += actual;
    }

  return res;
}

ssize_t
listxattr(const char *path, char *list, size_t size)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 1) == 0)
    return __xattr_filllist(&fib, list, size);
  return -1;
}

ssize_t
llistxattr(const char *path, char *list, size_t size)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 0) == 0)
    return __xattr_filllist(&fib, list, size);
  return -1;
}

ssize_t
flistxattr (int fd, char *list, size_t size)
{
  struct file *fp;
  struct FileInfoBlock fib;
  ssize_t res = -1;
  int omask;
  BPTR lock;
  usetup;

  if (fd >= NOFILE || (fp = u.u_ofile[fd]) == NULL)
    {
      errno = EBADF;
      return -1;
    }

  /* no extended attrs for null handler / sockets */
  if (HANDLER_NIL(fp) ||
      (fp->f_type == DTYPE_SOCKET && u.u_ixnetbase))
  {
    errno = ENOTSUP;
    return -1;
  }

  omask = syscall (SYS_sigsetmask, ~0);

  switch (fp->f_type)
    {
    case DTYPE_FILE:
      if (ExamineFH(CTOBPTR(fp->f_fh), &fib))
        res = __xattr_filllist(&fib, list, size);
      else
        errno = __ioerr_to_errno (IoErr());
      break;
    case DTYPE_MEM:
      if (fp->f_name)
        {
          lock = __lock(fp->f_name, ACCESS_READ);
          if (lock)
            {
              if (Examine(lock, &fib))
                res = __xattr_filllist(&fib, list, size);
              else
                errno = __ioerr_to_errno (IoErr());
              UnLock(lock);
            }
          else
            errno = __ioerr_to_errno (IoErr());
        }
      else
        errno = EINVAL;
      break;
    default:
      errno = ENOTSUP;
      break;
    }

  syscall (SYS_sigsetmask, omask);

  return res;
}
