/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *  Portions Copyright (C) 1994 Rafael W. Luebbert
 *  Portions Copyright (C) 1996 Jeff Shepherd
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Miscellaneous functions */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

pid_t
getpid(void)
{
#ifdef NATIVE_MORPHOS
  pid_t pid;
  NewGetTaskAttrsA(NULL, &pid, sizeof(pid), TASKINFOTYPE_PID, NULL);
  return pid;
#else
  return (pid_t)SysBase->ThisTask;
#endif
}


pid_t
getppid(void)
{
  usetup;
#ifdef NATIVE_MORPHOS
  pid_t pid;
  /* ix_open/ix_close use magic pid of "1" to indicate 'init'. This is somewhat
     wrong as this refers to CLI #1 under MorphOS. This may be a problem. - Piru
  */
  if (u.p_pptr == (struct Process *) 1)
    return 1;
  NewGetTaskAttrsA(&u.p_pptr->pr_Task, &pid, sizeof(pid), TASKINFOTYPE_PID, NULL);
  return pid;
#else
  return (int)u.p_pptr;
#endif
}

int
setpgid(pid_t pid, pid_t pgrp)
{
  usetup;

  if (pgrp < 0)
    {
      errno = EINVAL;
      return -1;
    }

  if (pid)
    {
#ifdef NATIVE_MORPHOS
      struct Task *task = FindTaskByPID(pid);
#else
      struct Task *task = (struct Task *) pid;
#endif
      struct user *u_ptr = task ? getuser(task) : NULL;
      if (!u_ptr)
	{
	  errno = ESRCH;
	  return -1;
	}
      u_ptr->p_pgrp = pgrp;
    }
  else
      u.p_pgrp = pgrp;
  return 0;
}

int
setpgrp(pid_t pid, pid_t pgrp)
{
  return setpgid(pid, pgrp);
}

pid_t
getpgrp(void)
{
  usetup;
  return u.p_pgrp;
}

pid_t
setsid(void)
{
  struct session *s;
  usetup;

  if (u.u_session && u.u_session->s_count <= 1)
    {
      errno = EPERM;
      return (pid_t)-1;
    }

  s = kmalloc(sizeof(struct session));
  if (s == NULL)
    {
      errno = ENOMEM;
      return (pid_t)-1;
    }
  if (u.u_session)
    u.u_session->s_count--;
  u.u_session = s;
  s->s_count = 1;
  s->pgrp = u.p_pgrp = getpid();

  return u.p_pgrp;
}

int
getrusage(int who, struct rusage *rusage)
{
  usetup;

  if (who != RUSAGE_SELF && who != RUSAGE_CHILDREN)
    {
      errno = EINVAL;
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      return -1;
    }

  *rusage = (who == RUSAGE_SELF) ? u.u_ru : u.u_cru;
  return 0;
}

char hostname[64] = "localhost";

int
gethostname(char *name, int namelen)
{
  usetup;

  if (u.u_ixnetbase)
    return netcall(NET_gethostname, name, namelen);
  strncpy (name, hostname, namelen);
  return 0;
}

int
sethostname(const char *name, int namelen)
{
  int len;
  usetup;

  if (u.u_ixnetbase)
    return netcall(NET_sethostname, name, namelen);

  len = namelen < sizeof (hostname) - 1 ? namelen : sizeof (hostname) - 1;
  strncpy (hostname, name, len);
  hostname[len] = 0;
  return 0;
}

/* not really useful.. but it's there ;-)) */
int
getpagesize(void)
{
#ifdef NATIVE_MORPHOS
  return page_size;
#else
  return 2048;
#endif
}

void sync (void)
{
  usetup;
  /* could probably walk the entire file table and fsync each, but
     this is too much of a nuisance ;-)) */
  errno = ENOSYS;
  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
}

int
fork (void)
{
  usetup;
  errno = ENOSYS;
  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
  return -1;
}

int
mkfifo (const char *path, mode_t mode)
{
  usetup;
  errno = ENOSYS;
  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
  return -1;
}

int
mknod (const char *path, mode_t mode, dev_t dev)
{
  int fd;
  usetup;

  switch (mode & S_IFMT)
    {
    case S_IFREG:
    case 0:
      break;
    default:
      errno = EINVAL;
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      return -1;
    }

  fd = syscall(SYS_open, path, O_WRONLY | O_CREAT | O_EXCL, mode);
  if (fd != -1)
    syscall(SYS_close, fd);

  return fd == -1 ? -1 : 0;
}

#if 0 //def NATIVE_MORPHOS
asm("
	.section \".text\"
	.globl	__flush_cache
	.type	__flush_cache,@function
__flush_cache:
	cmpwi	0,4,0
	addi	4,4,31
	mr	5,3
	srwi    4,4,5
	mtctr	4
	beqlr
l1:
	dcbf	0,3	/* flush data cache lines */
	addi	3,3,32
	bdnz	l1

	sync
	mtctr	4
l2:
	icbi	0,5	/* invalidate insn cache lines */
	addi	5,5,32
	bdnz	l2

	sync
	isync
	blr
__end_flush_cache:
	.size	__flush_cache,__end_flush_cache-__flush_cache
");

#else
void __flush_cache(void *addr, int len)
{
#ifdef NATIVE_MORPHOS
  CacheFlushDataInstArea(addr, len);
#endif
  CacheClearE(addr, len, CACRF_ClearI | CACRF_ClearD);
}
#endif


#if 0 //def NATIVE_MORPHOS
asm("
	.section \".text\"
	.globl	ix_flush_insn_cache
	.type	ix_flush_insn_cache,@function
ix_flush_insn_cache:
	cmpwi	0,4,0
	addi	4,4,31
	mr	5,3
	srwi    4,4,5
	mtctr	4
	beqlr
l3:
	dcbst	0,3	/* store dirty data cache lines */
	addi	3,3,32
	bdnz	l3

	sync
	mtctr	4
l4:
	icbi	0,5	/* invalidate insn cache lines */
	addi	5,5,32
	bdnz	l4

	sync
        isync
	blr
__end_ix_flush_insn_cache:
	.size	ix_flush_insn_cache,__end_ix_flush_insn_cache-ix_flush_insn_cache
");

#else
void ix_flush_insn_cache(void *addr, int len)
{
#ifdef NATIVE_MORPHOS
  CacheInvalidInstArea(addr, len);
#endif
  CacheClearE(addr, len, CACRF_ClearI);
}
#endif

#if 0 //def NATIVE_MORPHOS
asm("
	.section \".text\"
	.globl	ix_flush_data_cache
	.type	ix_flush_data_cache,@function
ix_flush_data_cache:
	cmpwi	0,4,0
	addi	4,4,31
	mr	5,3
	srwi    4,4,5
	mtctr	4
	beqlr
l5:
	dcbf    0,3	/* flush data cache lines */
	addi	3,3,32
	bdnz	l5

	sync
	blr
__end_ix_flush_data_cache:
	.size	ix_flush_data_cache,__end_ix_flush_data_cache-ix_flush_data_cache
");

#else
void ix_flush_data_cache(void *addr, int len)
{
#ifdef NATIVE_MORPHOS
  CacheFlushDataArea(addr, len);
  /*
   * Here we only flush data cache, so calling CacheClearE would do
   * nothing anyway. - Piru
   */
#else
  CacheClearE(addr, len, CACRF_ClearD);
#endif
}
#endif

void ix_flush_caches(void)
{
#ifdef NATIVE_MORPHOS
  CacheFlushDataInstArea(0, 0xffffffff);
#endif
  CacheClearU();
}
