/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  malloc.c,v 1.1.1.1 1994/04/04 04:30:29 amiga Exp
 *
 *  malloc.c,v
 * Revision 1.1.1.1  1994/04/04  04:30:29  amiga
 * Initial CVS check in.
 *
 *  Revision 1.5  1992/10/20  16:26:17  mwild
 *  change to single indirection to memory management list, allowing us
 *  to switch lists on the fly (important for vfork).
 *
 *  Revision 1.4  1992/09/14  01:44:41  mwild
 *  reduce magic cookie information by one long. By this, gcc's obstack will
 *  allocate in neat 4k blocks, which makes GigaMem happy (GigaMem is a commercial
 *  VMem package for AmigaOS).
 *
 *  Revision 1.3  1992/08/09  20:58:57  amiga
 *  add cast
 *
 *  Revision 1.2  1992/05/22  01:43:07  mwild
 *  use buddy-alloc memory management
 *
 * Revision 1.1  1992/05/14  19:55:40  mwild
 * Initial revision
 *
 */

#define _KERNEL
#include "ixemul.h"
#include "ixalloc.h"
#include "kprintf.h"

#include <stddef.h>
#include <unistd.h>

#ifndef USE_OLD_ALLOC

extern void *ix_pub_mempool;

#define USE_PANIC_HIT 0

#define mem_list (u.u_mdp->md_list)
#define mem_used (u.u_mdp->md_malloc_sbrk_used)
#define own_mem_list (u.u_md.md_list)
#define own_mem_used (u.u_md.md_malloc_sbrk_used)

#define MAGIC1 0x55555555
#define MAGIC2 0xaaaaaaaa

/* use only one magic cookie at the end, since then the most frequently
   done allocation of gcc, 4072 byte, fits neatly into 4096 when we've added
   our information. Gigamem will thank you that it doesn't have to give
   you an additional page for 4 byte ;-)) */

typedef struct  {
  struct ixnode    node;
  u_int            size;
  u_int            magic[2];
  /* Please note there is also a magic2[1] at the end of allocated block! */
} __malloc_t;

#define MALLOC_HEADER_SIZE     ((sizeof(__malloc_t) + MEM_BLOCKMASK) & ~MEM_BLOCKMASK)
#define MALLOC_HEADER_2MEM(_h) ((void *)(((u_char *)(_h)) + MALLOC_HEADER_SIZE))
#define MALLOC_MEM_2HEADER(_m) ((__malloc_t *)(((u_char *)(_m)) - MALLOC_HEADER_SIZE))

static inline __malloc_t *
b_alloc (size_t size, int align)
{
	__malloc_t *res;
	usetup;
	Forbid();
	res = AllocPooledAligned(ix_pub_mempool, size + MALLOC_HEADER_SIZE + sizeof(u_int), align, MALLOC_HEADER_SIZE); /* Include one additional sizeof(u_int) for magic2[1] */
	if (res)
		mem_used += size;
	Permit();
	return res;
}

static inline void
b_free (__malloc_t *block)
{
	usetup;
	size_t size = block->size;
	Forbid();
	FreePooled(ix_pub_mempool, block, size + MALLOC_HEADER_SIZE + sizeof(u_int)); /* Include one additional sizeof(u_int) for magic2[1] */
	mem_used -= size;
	Permit();
}

static void *
ix_malloc (size_t size, int align)
{
	__malloc_t *res;
	int omask;
	usetup;

	/* We increase SIZE below which could cause an dangerous (system
	   crash) overflow. -bw/09-Jun-98 */
	if((signed long)size >= 0)
	{
		/* guarantee long sizes (so we can use CopyMemQuick in realloc) */
		size = (size + 3) & ~3; /* next highest multiple of 4 */

		/* we don't want to be interrupted between the allocation and the tracking */
		omask = syscall (SYS_sigsetmask, ~0);

		/* include management information */
		res = b_alloc(size, align);
		if (res)
		{
			/* Setup header */
			res->size = size;
			res->magic[0] = MAGIC1;
			res->magic[1] = MAGIC1;
			/* MAGIC2 at the end of allocated block */
			*((u_int *)((u_char *)MALLOC_HEADER_2MEM(res) + size)) = MAGIC2;

			Forbid ();
			ixaddtail ((struct ixlist *)&mem_list, (struct ixnode *)res);
			Permit ();

			syscall (SYS_sigsetmask, omask);

			return MALLOC_HEADER_2MEM(res);
		}

		syscall (SYS_sigsetmask, omask);
	}
	errno = ENOMEM;
	return NULL;
}

void *
malloc (size_t size)
{
	/* Wrap to ix_malloc() with default alignment */
	return ix_malloc(size, IXALLOC_ALIGN);
}

void *
memalign (size_t alignment, size_t size)
{
	return ix_malloc(size, alignment);
}

void *
valloc (size_t size)
{
	/* Same as memalign but aligned to PAGE_SIZE */
	return ix_malloc(size, page_size);
}

static inline void
check_header(const __malloc_t *block, const char __func[])
{
	/* MAGIC1 check */
	if(((u_int)block & 1) ||
		block->magic[0] != MAGIC1 || block->magic[1] != MAGIC1)
	{
#if USE_PANIC_HIT
		*(int*)0=0;
#endif
		ix_panic ("%s: start of block corrupt!", __func);
		syscall (SYS_exit, 20);
	}

	/* MAGIC2 check */
	if(*((u_int *)((u_char *)MALLOC_HEADER_2MEM(block) + block->size)) != MAGIC2)
	{
#if USE_PANIC_HIT
		*(int*)0=0;
#endif
		ix_panic ("%s: end of block corrupt!", __func);
		syscall (SYS_exit, 20);
	}
}

void
free (void *mem)
{
	__malloc_t *block;
	int omask;
	usetup;

	/* this seems to be standard... don't like it though ! */
	if (! mem)
		return;

	block = MALLOC_MEM_2HEADER(mem);

	/* Check memory block */
	check_header(block, __FUNCTION__);

	/* we don't want to be interrupted between unlinking the block and freeing it */
	omask = syscall (SYS_sigsetmask, ~0);

	Forbid ();
	ixremove ((struct ixlist *)&mem_list, (struct ixnode *) block);
	Permit ();

	b_free(block);

	syscall (SYS_sigsetmask, omask);
}

void all_free (void)
{
	__malloc_t *b, *nb;
	static int errno_dummy;
	usetup;

	/* Since errno (= *u.u_errno) might still be used, we point it to our own
	   temporary errno dummy variable. */
	u.u_errno = &errno_dummy;

#ifdef NATIVE_MORPHOS
	/* If running modern exec call the TLS destructors (if any) before
	   releasing our memory. This tries to ensure that we can't have a a
	   situation where the destructor memory is already free. - Piru */
	if (LIB_MINVER(&SysBase->LibNode, 51, 46))
	{
		TLSCallDestructors(NULL);
	}
#endif

	/* Atempt to cleanup previous __alloca() resources */
	if(u.u_alloca != NULL)
		__alloca_cleanup();

	/* be sure to use *our* memory lists in here, never all_free() the parents
	   list.... */
	for (b = (__malloc_t *)own_mem_list.head; b; b = nb)
	{
		nb = (__malloc_t *)b->node.next;
		b_free(b);
	}

	/* this makes it possible to call all_free() more than once */
	ixnewlist ((struct ixlist *)&own_mem_list);
}

void *
realloc (void *old_mem, size_t new_size)
{
	__malloc_t *block;
	void *res = NULL;
	usetup;

	if(!old_mem)
		return malloc(new_size);

	if(!new_size)
	{
		free(old_mem);
		return NULL;
	}

	block = MALLOC_MEM_2HEADER(old_mem);

	/* Check memory block */
	check_header(block, __FUNCTION__);

	/* now that the block is validated, check whether we have to really
	 * realloc, or if we just can return the old block */
	if(block->size >= new_size)
		return old_mem;

	res = malloc(new_size);
	if (res)
	{
		CopyMemQuick (old_mem, res, block->size);

		/* according to the manpage, the old buffer should only be
		 * freed, if the allocation of the new buffer was successful */
		free(old_mem);
	}

	return res;
}

#endif
