/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 2020  Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  pwrite64.c,v 1.1.1.1 2020/01/15 00:30:30 amiga Exp
 *
 *  Revision 1.1  2020/01/15  00:30:30  piru
 *  Initial revision
 *
 */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"
#include <unistd.h>

int64_t pwrite64(int fd, const void *buf, u_int64_t len, off64_t off)
{
  usetup;
  struct file *f;
  int64_t result = -1LL;

  if (buf == NULL || ((u_int64_t) buf + len) > 0xffffffffULL)
    {
      errno = EFAULT;
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
    }
  /* if this is an open fd */
  else if (fd >= 0 && fd < NOFILE && (f = u.u_ofile[fd]))
    {
      if (f->f_type == DTYPE_FILE)
	{
	  const int doseek = !HANDLER_NIL(f);

	  if (!doseek && off != 0LL)
	    {
	      errno = ESPIPE;
	      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	    }
	    else if (f->f_write)
	    {
	      int omask = 0;
	      QUAD oldposition = 0LL;

	      if (doseek)
		{
		  omask = syscall (SYS_sigsetmask, ~0);
		  __get_file (f);

		  oldposition = Seek64(CTOBPTR(f->f_fh), 0, OFFSET_CURRENT);
		  if (oldposition != -1LL)
		    {
		      /* Need to call lseek64 so the file is expanded if seekpos is past EOF */
		      if (lseek64(fd, off, SEEK_SET) != off)
			{
			  oldposition = -1LL;
			}
		    }
		}
	      if (oldposition != -1LL)
		{
		  // __release_file (f);
		  result = (*f->f_write)(f, buf, (size_t) len);
		  //__get_file (f);

		  if (result != -1LL && result != 0LL)
		    f->f_stb_dirty &= ~FSDF_UTIME;

		  if (doseek && Seek64(CTOBPTR(f->f_fh), oldposition, OFFSET_BEGINNING) == -1LL)
		    {
		      errno = EIO;
		      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
		      result = -1LL;
		    }
		}
	      else
		{
		  errno = EIO;
		  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
		}

	      if (doseek)
		{
		   __release_file (f);
		  syscall (SYS_sigsetmask, omask);
		}
	    }
          else
	    {
	      errno = EIO;
	      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	   }
	}
      else if (f->f_type == DTYPE_MEM)
	{
	  if (f->f_write)
	    {
	      off64_t oldposition;
	      int omask = syscall (SYS_sigsetmask, ~0);
	      __get_file (f);

	      oldposition = lseek64(fd, SEEK_CUR, 0);
	      if (oldposition != -1LL)
		{
		  if (lseek64(fd, SEEK_SET, off) == off)
		    {
		      result = (*f->f_write)(f, buf, (size_t) len);
		      if (lseek64(fd, SEEK_SET, oldposition) != oldposition)
		        {
		          result = -1LL;
		        }
		    }
		}

	      __release_file (f);
	      syscall (SYS_sigsetmask, omask);
	    }
	  else
	    {
	      errno = EIO;
	    }
	}
      else
	{
	  errno = EPIPE;
	}
    }
  else
    {
      errno = EBADF;
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
    }

  if (result != -1LL)
    errno = 0;

  return result;
}

ssize_t pwrite(int fd, const void *buf, size_t len, off_t off)
{
  return (ssize_t) pwrite64(fd, buf, len, off);
}
