/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 2015  Harry Sintonen.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define _KERNEL
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "ixemul.h"
#include "kprintf.h"

extern const char SYS_XATTR_AMIGA_MODE[];
extern const char SYS_XATTR_AMIGA_COMMENT[];

extern int
__xattr_examine(const char *path, struct FileInfoBlock *fib, int followsoftlink);

extern char *ix_to_ados(char *, const char *);

static ssize_t
__xattr_removeattr(const char *path, struct FileInfoBlock *fib, const char *name)
{
  usetup;

  if (!strcmp(name, SYS_XATTR_AMIGA_MODE))
    {
      /* Note: protection always exists, simulate success */
      return 0;
    }
  else if (!strcmp(name, SYS_XATTR_AMIGA_COMMENT))
    {
      char *adospath;
      int omask;
      int res;
      if (fib->fib_Comment[0] == '\0')
        {
          errno = ENOATTR;
          return -1;
        }
      adospath = malloc(strlen(path) + 4);
      if (adospath == 0)
        {
          errno = ENOMEM;
          return -1;
        }
      ix_to_ados(adospath, path);
      omask = syscall (SYS_sigsetmask, ~0);
      res = SetComment(adospath, "") == 0 ? -1 : 0;
      free(adospath);
      if (res == -1)
        __ioerr_to_errno (IoErr());

      syscall (SYS_sigsetmask, omask);
      return res;
    }

  errno = ENOATTR;
  return -1;
}

ssize_t
removexattr(const char *path, const char *name)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 1) == 0)
    return __xattr_removeattr(path, &fib, name);
  return -1;
}

ssize_t
lremovexattr(const char *path, const char *name)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 0) == 0)
    return __xattr_removeattr(path, &fib, name);
  return -1;
}

ssize_t
fremovexattr (int fd, const char *name)
{
  struct file *fp;
  struct FileInfoBlock fib;
  ssize_t res = -1;
  int omask;
  BPTR lock;
  usetup;

  if (fd >= NOFILE || (fp = u.u_ofile[fd]) == NULL)
    {
      errno = EBADF;
      return -1;
    }

  /* no external attrs for null handler / sockets */
  if (HANDLER_NIL(fp) ||
      (fp->f_type == DTYPE_SOCKET && u.u_ixnetbase))
    {
      errno = ENOTSUP;
      return -1;
    }

  omask = syscall (SYS_sigsetmask, ~0);

  switch (fp->f_type)
    {
    case DTYPE_FILE:
      if (fp->f_parentlock && fp->f_name)
        {
          if (ExamineFH(CTOBPTR(fp->f_fh), &fib))
            {
              BPTR olddir = CurrentDir(fp->f_parentlock);
              res = __xattr_removeattr(fp->f_name, &fib, name);
              CurrentDir(olddir);
            }
          else
           errno = __ioerr_to_errno (IoErr());
        }
      else
        errno = EINVAL;
      break;
    case DTYPE_MEM:
      if (fp->f_name)
        {
          lock = __lock(fp->f_name, ACCESS_READ);
          if (lock)
            {
              if (Examine(lock, &fib))
                res = __xattr_removeattr(fp->f_name, &fib, name);
              else
                errno = __ioerr_to_errno (IoErr());
              UnLock(lock);
            }
          else
            errno = __ioerr_to_errno (IoErr());
        }
      else
        errno = EINVAL;
      break;
    default:
      errno = ENOTSUP;
      break;
    }

  syscall (SYS_sigsetmask, omask);

  return res;
}
