/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *  Copyright (C) 2020  Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: rename.c,v 1.5 2021/02/21 21:21:32 piru Exp $
 */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"
#include <stdlib.h>

#if 0

struct rename_vec {
  char *source_name;
  BPTR target_dir_lock;
  BSTR target_name;
};


static int
__rename_func (struct lockinfo *info, struct rename_vec *rv, int *error)
{
  struct StandardPacket *sp = &info->sp;

  if (SameLock(info->parent_lock, rv->target_dir_lock) == LOCK_DIFFERENT)
    {
      sp->sp_Pkt.dp_Res1 = 0;
      sp->sp_Pkt.dp_Res2 = ERROR_RENAME_ACROSS_DEVICES;
    }
  else
    {
      sp->sp_Pkt.dp_Type = ACTION_RENAME_OBJECT;
      sp->sp_Pkt.dp_Arg1 = info->parent_lock;
      sp->sp_Pkt.dp_Arg2 = info->bstr;
      sp->sp_Pkt.dp_Arg3 = rv->target_dir_lock;
      sp->sp_Pkt.dp_Arg4 = rv->target_name;

      PutPacket (info->handler, sp);
      __wait_sync_packet (sp);
    }

  info->result = sp->sp_Pkt.dp_Res1;
  *error = info->result != -1;

  /* retry if necessary */
  return 1;
}

static int
__get_target_data (struct lockinfo *info, struct rename_vec *rv, int *error)
{
  struct StandardPacket *sp = &info->sp;

  rv->target_dir_lock = info->parent_lock;
  rv->target_name = info->bstr;

  info->result = __plock (rv->source_name, __rename_func, rv);
  sp->sp_Pkt.dp_Res2 = IoErr();

  *error = info->result != -1;

  /* don't retry */
  return 0;
}

int
rename(const char *from, const char *to)
{
  int res;
  struct rename_vec rv;
  usetup;

  rv.source_name = (char *)from;

  res = __plock ((char *)to, __get_target_data, &rv);

  if (res == 0 && IoErr() == ERROR_OBJECT_EXISTS)
    {
      syscall (SYS_unlink, to);
      res = __plock ((char *)to, __get_target_data, &rv);
    }

  if (!res)
    errno = __ioerr_to_errno (IoErr());

  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
  return res ? 0 : -1;
}

#else

char *ix_to_ados(char *, const char *);
char *check_root(char *);

static
LONG try_softlink_rename(char *from, char *to, int softlinkindex);

int
rename(const char *from, const char *to)
{
  int res;
  usetup;
  int omask;
  char *buf1 = alloca(strlen(from) + 4);
  char *buf2 = alloca(strlen(to) + 4);
  LONG err = 0;
  struct FileInfoBlock fib;
  int dstisdir = 0;
  int dstissoftlink = 0;
  int dstisemptydir = 0;
  BPTR lock;
  char *tmpname = NULL;

  buf1 = ix_to_ados(buf1, from);
  buf2 = ix_to_ados(buf2, to);

  omask = syscall (SYS_sigsetmask, -1);

  /* If target exists, source need to be of same type.
     If target is a directory, the directory must be empty.
     We need to be able to replace the target if it exists. */
  lock = __llock (to, ACCESS_READ);
  if (lock)
    {
      if (Examine (lock, &fib))
        {
          dstisdir = fib.fib_DirEntryType > 0;
          dstisemptydir = dstisdir && !ExNext (lock, &fib) && IoErr () == ERROR_NO_MORE_ENTRIES;
        }
      __unlock (lock);
    }
  else if (IoErr () == ERROR_IS_SOFT_LINK)
    {
      /* Target is a soft link. */
      dstissoftlink = 1;
    }

  /* If target exists */
  if (lock || dstissoftlink)
    {
      int srcnotdir = 0;
      lock = __llock (from, ACCESS_READ);
      if (lock)
        {
          if (dstissoftlink)
            {
              /* Target is a softlink but source is not. */
              __unlock(lock);
              syscall (SYS_sigsetmask, omask);
              errno = ENOTDIR;
              return -1;
            }
          if (Examine (lock, &fib))
            srcnotdir = fib.fib_DirEntryType < 0;
          __unlock(lock);
        }
      else if (IoErr () == ERROR_IS_SOFT_LINK)
        srcnotdir = 1; /* is a softlink, so not a dir */

      if (dstisdir == srcnotdir)
        {
          /* Target and source are of different type. */
          syscall (SYS_sigsetmask, omask);
          errno = dstisdir ? EISDIR : ENOTDIR;
          return -1;
        }
      if (dstisdir && !dstisemptydir)
        {
          /* Both source and target are directories but the target
             directory is not empty. */
          syscall (SYS_sigsetmask, omask);
          errno = ENOTEMPTY;
          return -1;
        }

      /* Rename the target object out of the way so that Rename() will
         do what we expect it to. If it wouldn't do this it would result
         in moving the source object inside the destination if target is a
         directory, or failing to rename if target is a file/softlink
         If the final Rename() works we can then really delete the old
         object. If the operaion fails, we can back out by renaming the
         object back. */
      tmpname = malloc (strlen(buf2) + 3 + 11 + 1);
      if (tmpname)
        {
          ULONG pid;
#ifdef NATIVE_MORPHOS
          NewGetTaskAttrsA (NULL, &pid, sizeof(pid), TASKINFOTYPE_PID, NULL);
#else
          pid = (ULONG) SysBase->ThisTask;
#endif
          sprintf (tmpname, "%s.t.%lu", buf2, pid);
        }
      if (!tmpname || !Rename(buf2, tmpname))
        {
          syscall (SYS_sigsetmask, omask);
          if (tmpname)
            {
              free (tmpname);
              errno = __ioerr_to_errno (IoErr ());
            }
          return -1;
        }
    }

  res = Rename(buf1, buf2);

  if (!res)
    {
      err = IoErr();
      if (err == ERROR_IS_SOFT_LINK)
	{
	  /* It's only safe to try this workaround if source and/or target is
	     absolute. */
	  BOOL srcabsolute = PathPart(buf1) != (STRPTR) buf1;
	  BOOL dstabsolute = PathPart(buf2) != (STRPTR) buf2;
	  if (srcabsolute || dstabsolute)
	    {
	      /* First if dst is absolute, assume source had a softlink */
	      if (dstabsolute)
		res = try_softlink_rename(buf1, buf2, 0);
	      /* If still failed with ERROR_IS_SOFT_LINK, and src is absolute,
	         assume dst has a softlink. */
	      if (!res && IoErr() == ERROR_IS_SOFT_LINK && srcabsolute)
		res = try_softlink_rename(buf1, buf2, 1);
	    }
	  else
	    {
	      /* If both are relative we're out of luck. We could try Lock()
	         both, NameFromLock() both and try Rename() but this also
	         has potential complications. We fail in this case for now.
	         - Piru */
	    }
	}
      else if (err == ERROR_OBJECT_NOT_FOUND)
	{
	  char *new_buf1 = check_root(buf1);
	  char *new_buf2 = check_root(buf2);
	  int retry = 0;
	  if (new_buf1 && *new_buf1)
	    {
	      buf1 = new_buf1;
	      retry = 1;
	    }
	  if (new_buf2 && *new_buf2)
	    {
	      buf2 = new_buf2;
	      retry = 1;
	    }
	  if (retry)
	    {
	      res = Rename(buf1, buf2);
	      if (!res)
		err = IoErr();
	    }
	}
    }

  if (tmpname)
    {
      if (res)
        {
          /* Try to delete the original target. */
          if (!DeleteFile (tmpname) && IoErr () == ERROR_DELETE_PROTECTED &&
              SetProtection (tmpname, 0))
            DeleteFile (tmpname);
        }
      else
        {
          /* Best effort trying to restore the original target. */
          err = IoErr();
          Rename (tmpname, buf2);
          SetIoErr (err);
        }
      free (tmpname);
    }

  syscall (SYS_sigsetmask, omask);

  if (!res)
    errno = __ioerr_to_errno (IoErr());

  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
  return res ? 0 : -1;
}

static
LONG try_softlink_rename(char *from, char *to, int softlinkindex)
{
  char *softlink = softlinkindex ? to : from;
  char *pathpos = PathPart(softlink);
  LONG result = DOSFALSE;
  char oldch;
  BPTR dirlock;

  if (softlinkindex)
    to = FilePart(to);
  else
    from = FilePart(from);

  oldch = *pathpos;
  *pathpos = '\0';
  dirlock = Lock(softlink, ACCESS_READ);
  *pathpos = oldch;
  if (dirlock)
    {
      dirlock = CurrentDir(dirlock);
      result = Rename(from, to);
      UnLock(CurrentDir(dirlock));
    }
  return result;
}

#endif
