/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 2015-2020  Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define _KERNEL
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "ixemul.h"
#include "kprintf.h"

#include <sys/xattr.h>

extern const char SYS_XATTR_AMIGA_MODE[];
extern const char SYS_XATTR_AMIGA_COMMENT[];

extern int
__xattr_examine(const char *path, struct FileInfoBlock *fib, int followsoftlink);

extern char *ix_to_ados(char *, const char *);

static ssize_t
__xattr_setattr(const char *path, struct FileInfoBlock *fib, const char *name, const void *value, size_t size, int flags)
{
  int res = -1;
  char *adospath;
  int omask;
  size_t actual;
  usetup;

  adospath = malloc(strlen(path) + 4);
  if (adospath == 0)
    {
      errno = ENOMEM;
      return -1;
    }
  ix_to_ados(adospath, path);

  omask = syscall (SYS_sigsetmask, ~0);

  if (!strcmp(name, SYS_XATTR_AMIGA_MODE))
    {
      /* Note: protection always exists */
      if (flags & XATTR_CREATE)
        {
          errno = EEXIST;
          goto end;
        }
      actual = 4;
      if (size >= actual)
        {
          const uint8_t *data = value;
          ULONG mode = (data[0] << 24) | (data[1] << 16) | (data[2] << 8) | data[3];
          res = SetProtection(adospath, mode);
          if (res == 0 && IoErr() == ERROR_IS_SOFT_LINK)
            {
              char *pathpos = PathPart(adospath);
              if (pathpos != adospath) /* There must be a dir component */
                {
                  BPTR dirlock;
                  char oldch = *pathpos;
                  *pathpos = '\0';
                  dirlock = Lock(adospath, ACCESS_READ);
                  if (dirlock)
                    {
                      dirlock = CurrentDir(dirlock);
                      *pathpos = oldch;
                      res = SetProtection(FilePart(adospath), mode);
                      UnLock(CurrentDir(dirlock));
                    }
                }
            }
          res = res == DOSFALSE ? -1 : 0;
          if (res == -1)
            __ioerr_to_errno (IoErr());
        }
      else
        errno = ERANGE;
    }
  else if (!strcmp(name, SYS_XATTR_AMIGA_COMMENT))
    {
      if ((flags & XATTR_CREATE) && fib->fib_Comment[0] != '\0')
        {
          errno = EEXIST;
          goto end;
        }
      if ((flags & XATTR_REPLACE) && fib->fib_Comment[0] == '\0')
        {
          errno = ENOATTR;
          goto end;
        }
      if (size > 0 && size <= 80 && ((char *) value)[size - 1] == '\0')
        {
          res = SetComment(adospath, (STRPTR) value);
          if (res == 0 && IoErr() == ERROR_IS_SOFT_LINK)
            {
              char *pathpos = PathPart(adospath);
              if (pathpos != adospath) /* There must be a dir component */
                {
                  BPTR dirlock;
                  char oldch = *pathpos;
                  *pathpos = '\0';
                  dirlock = Lock(adospath, ACCESS_READ);
                  if (dirlock)
                    {
                      dirlock = CurrentDir(dirlock);
                      *pathpos = oldch;
                      res = SetComment(FilePart(adospath), (STRPTR) value);
                      UnLock(CurrentDir(dirlock));
                    }
                }
            }
          res = res == DOSFALSE ? -1 : 0;
          if (res == -1)
            __ioerr_to_errno (IoErr());
        }
      else
        errno = ERANGE;
    }
  else
    errno = ENOTSUP;

  end:
  syscall (SYS_sigsetmask, omask);

  free(adospath);

  return res;
}

ssize_t
setxattr(const char *path, const char *name, void *value, size_t size, int flags)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 1) == 0)
    return __xattr_setattr(path, &fib, name, value, size, flags);
  return -1;
}

ssize_t
lsetxattr(const char *path, const char *name, void *value, size_t size, int flags)
{
  struct FileInfoBlock fib;
  if (__xattr_examine(path, &fib, 0) == 0)
    return __xattr_setattr(path, &fib, name, value, size, flags);
  return -1;
}

ssize_t
fsetxattr (int fd, const char *name, void *value, size_t size, int flags)
{
  struct file *fp;
  struct FileInfoBlock fib;
  ssize_t res = -1;
  int omask;
  BPTR lock;
  usetup;

  if (fd >= NOFILE || (fp = u.u_ofile[fd]) == NULL)
    {
      errno = EBADF;
      return -1;
    }

  /* no external attrs for null handler / sockets */
  if (HANDLER_NIL(fp) ||
      (fp->f_type == DTYPE_SOCKET && u.u_ixnetbase))
    return 0;

  omask = syscall (SYS_sigsetmask, ~0);

  switch (fp->f_type)
    {
    case DTYPE_FILE:
      if (fp->f_parentlock && fp->f_name)
        {
          if (ExamineFH(CTOBPTR(fp->f_fh), &fib))
            {
              BPTR olddir = CurrentDir(fp->f_parentlock);
              res = __xattr_setattr(fp->f_name, &fib, name, value, size, flags);
              CurrentDir(olddir);
            }
          else
            errno = __ioerr_to_errno (IoErr());
        }
      else
        errno = EINVAL;
      break;
    case DTYPE_MEM:
      if (fp->f_name)
        {
          lock = __lock(fp->f_name, ACCESS_READ);
          if (lock)
            {
              if (Examine(lock, &fib))
                res = __xattr_setattr(fp->f_name, &fib, name, value, size, flags);
             else
                errno = __ioerr_to_errno (IoErr());
              UnLock(lock);
            }
          else
            errno = __ioerr_to_errno (IoErr());
        }
      else
        errno = EINVAL;
      break;
    default:
      errno = ENOTSUP;
      break;
    }

  syscall (SYS_sigsetmask, omask);

  return res;
}
