/*
 *  This file is part of ixemul.library for MorphOS.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *  Copyright (C) 2015  Harry Sintonen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define _KERNEL
#include "ixemul.h"
#include "kprintf.h"
#include <sys/mount.h>
#include <string.h>

/* no comments here ;-) */
#define ID_ALT_OFS_DISK         (0x444F5304L)   /* 'DOS\4' */
#define ID_ALT_FFS_DISK         (0x444F5305L)   /* 'DOS\5' */
#define ID_ALT_INTER_DOS_DISK   (0x444F5306L)   /* 'DOS\6' */
#define ID_ALT_INTER_FFS_DISK   (0x444F5307L)   /* 'DOS\7' */

/* dunno, would be logical ;-)) */
#define ID_NFS_DISK             (0x4E465300)    /* 'NFS\0' */
/* did somebody already do this ??? */
#define ID_UFS_DISK             (0x55465300)    /* 'UFS\0' */


static int safedopkt3(struct statfs64 *buf, LONG action, LONG a1, LONG a2, LONG a3)
{
  usetup;

  struct DosList *dl;
  struct StandardPacket *sp;

  sp = alloca (sizeof (*sp) + 2);
  sp = LONG_ALIGN (sp);
  __init_std_packet (sp);

  /*
   * Make sure the handler process is still valid before attempting
   * to send dospacket to it.
   */

  dl = LockDosList(LDF_DEVICES | LDF_READ);
  while ((dl = NextDosEntry(dl, LDF_DEVICES | LDF_READ)))
    {
      if ((struct MsgPort *) buf->f_fsid.val[0] == dl->dol_Task)
	{
	  sp->sp_Pkt.dp_Port = u.u_sync_mp;
	  sp->sp_Pkt.dp_Type = action;
	  sp->sp_Pkt.dp_Arg1 = a1;
	  sp->sp_Pkt.dp_Arg2 = a2;
	  sp->sp_Pkt.dp_Arg3 = a3;
	  /*
	   * The forbid here makes sure the packet isn't really
	   * processed while we're still holding the doslock.
	   */
	  Forbid();
	  PutPacket (dl->dol_Task, sp);
	  UnLockDosList(LDF_DEVICES | LDF_READ);
	  Permit();
	  break;
	}
    }

  if (!dl)
    {
      /* The device could not be found anymore, back out! */
      UnLockDosList(LDF_DEVICES | LDF_READ);
      return 0;
    }

  __wait_sync_packet (sp);

  return sp->sp_Pkt.dp_Res1;
}


/* this function fills in InfoData information from a statfs64 structure
   that has at least its f_fsid field set to the handler */
static void
internal_statfs64 (struct statfs64 *buf)
{
  struct InfoData *info;
  UQUAD numblocks, numblocksused;

  info = alloca (sizeof (*info) + 2);
  info = LONG_ALIGN (info);

  if (!safedopkt3(buf, ACTION_DISK_INFO, CTOBPTR (info), 0, 0))
    {
      /* The device could not be found anymore or some other error, back out! */
      return;
    }

#ifdef NATIVE_MORPHOS
  if (!safedopkt3(buf, ACTION_QUERY_ATTR, FQA_NumBlocks, (LONG) &numblocks, sizeof(numblocks)) ||
      !safedopkt3(buf, ACTION_QUERY_ATTR, FQA_NumBlocksUsed, (LONG) &numblocksused, sizeof(numblocksused)))
    {
      /* Either device is gone or the FS doesn't support ACTION_QUERY_ATTR.
         Either way, use values gotten from ACTION_DISK_INFO. */
      numblocks = (ULONG) info->id_NumBlocks;
      numblocksused = (ULONG) info->id_NumBlocksUsed;
    }
#endif

  /* packet was implemented on handler side */
    {
      struct DeviceList *dl = BTOCPTR (info->id_VolumeNode);

      if (dl)
	{
	  unsigned char *name = BTOCPTR (dl->dl_Name);

	  /* use this field to represent the volume name */
	  buf->f_mntfromname[0] = '/';
	  bcopy (name + 1, buf->f_mntfromname + 1, name[0]);
	  buf->f_mntfromname[name[0] + 1] = 0;
	}
      else
	/* for stuff like FIFO that doesn't have an associated volume */
	strcpy (buf->f_mntfromname, buf->f_mntonname);

      switch (info->id_DiskType)
	{
	case ID_DOS_DISK:
	case ID_ALT_OFS_DISK:
	  buf->f_type = MOUNT_ADOS_OFS;
	  break;

	case ID_FFS_DISK:
	case ID_ALT_FFS_DISK:
	  buf->f_type = MOUNT_ADOS_FFS;
	  break;

	case ID_INTER_DOS_DISK:
	case ID_ALT_INTER_DOS_DISK:
	  buf->f_type = MOUNT_ADOS_IOFS;
	  break;

	case ID_INTER_FFS_DISK:
	case ID_ALT_INTER_FFS_DISK:
	  buf->f_type = MOUNT_ADOS_IFFS;
	  break;

	case ID_MSDOS_DISK:
	  buf->f_type = MOUNT_PC;
	  break;

	case ID_NFS_DISK:
	  buf->f_type = MOUNT_NFS;
	  break;

	case ID_UFS_DISK:
	  buf->f_type = MOUNT_UFS;
	  break;

	default:
	  buf->f_type = MOUNT_NONE;
	  break;
	}

      buf->f_flags  = MNT_NOSUID | MNT_NODEV;
      if (info->id_DiskState != ID_VALIDATED)
	/* can be ID_WRITE_PROTECTED or ID_VALIDATING */
	buf->f_flags |= MNT_RDONLY;

      buf->f_fsize = info->id_BytesPerBlock;
      buf->f_bsize = info->id_BytesPerBlock * ix.ix_fs_buf_factor;
      buf->f_blocks = numblocks;
      buf->f_bfree =
	buf->f_bavail = numblocks - numblocksused;
      /* no inode information available, thus set to -1 */
      buf->f_files =
      buf->f_ffree = -1;
   }
}


int
getfsstat64 (struct statfs64 *buf, long bufsize, int flags)
{
  int num_devs = 0;
  struct statfs64 *orig_buf = buf;
  long orig_bufsize = bufsize;
  struct DevInfo *dv;

  dv = (APTR) LockDosList(LDF_DEVICES | LDF_READ);
  while ((dv = (APTR) NextDosEntry((APTR) dv, LDF_DEVICES | LDF_READ)))
    {
      if (dv->dvi_Task && dv->dvi_Name)
	{
	  num_devs ++;
	  if (buf && bufsize >= sizeof (*buf))
	    {
	      unsigned char *name = BTOCPTR (dv->dvi_Name);

	      /* only remember the name and the address of the
		 handler so that we can later send INFO packets there */
	      bzero (buf, sizeof (*buf));
	      buf->f_mntonname[0] = '/';
	      bcopy (name + 1, buf->f_mntonname + 1, name[0]);
	      //buf->f_mntonname[name[0] + 1] = 0;
	      //buf->f_fsid.val[1] = 0;
	      buf->f_fsid.val[0] = (long) dv->dvi_Task;

	      buf++;
	      bufsize -= sizeof (*buf);
	    }
	}
    }
  UnLockDosList(LDF_DEVICES | LDF_READ);

  if (! orig_buf || orig_bufsize < sizeof (*buf))
    return num_devs;

  /* have a second run, and inquire more data from the associated volumes */
  while (orig_buf < buf)
    {
      internal_statfs64 (orig_buf);
      orig_buf++;
    }

  return num_devs;
}

int
fstatfs64 (int fd, struct statfs64 *buf)
{
  usetup;
  struct file *f;
  struct DevInfo *dv;

  if ((unsigned)fd < NOFILE && (f = u.u_ofile[fd]))
    {
      if (! buf)
	{
	  errno = EFAULT;
	  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	  return -1;
	}

      if (HANDLER_NIL (f) || f->f_type != DTYPE_FILE)
	{
	  errno = EOPNOTSUPP;
	  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	  return -1;
	}

      bzero (buf, sizeof (*buf));
      //buf->f_fsid.val[1] = 0;
      buf->f_fsid.val[0] = (long) f->f_fh->fh_Type;

      dv = (APTR) LockDosList(LDF_DEVICES | LDF_READ);
      while ((dv = (APTR) NextDosEntry((APTR) dv, LDF_DEVICES | LDF_READ)))
	{
	  if ((struct MsgPort *) dv->dvi_Task == f->f_fh->fh_Type)
	    {
	      unsigned char *name = BTOCPTR (dv->dvi_Name);

	      buf->f_mntonname[0] = '/';
	      bcopy (name + 1, buf->f_mntonname + 1, name[0]);
	      //buf->f_mntonname[name[0] + 1] = 0;
	      break;
	    }
	}
      UnLockDosList(LDF_DEVICES | LDF_READ);

      if (! dv)
	{
	  errno = EOPNOTSUPP;
	  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	  return -1;
	}

      internal_statfs64 (buf);

      return 0;
    }

  errno = EBADF;
  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
  return -1;
}


int
statfs64 (const char *path, struct statfs64 *buf)
{
  usetup;
  struct MsgPort *handler;
  BPTR lock;
  int omask, err;
  struct DevInfo *dv;

  omask = syscall (SYS_sigsetmask, ~0);
  lock = __lock ((char *)path, ACCESS_READ);
  handler = lock ? ((struct FileLock *) BTOCPTR (lock))->fl_Task : 0;
  err = __ioerr_to_errno (IoErr ());
  if (lock)
    __unlock (lock);
  syscall (SYS_sigsetmask, omask);

  if (! handler)
    {
      errno = err;
      KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
      return -1;
    }
  else
    {
      if (! buf)
	{
	  errno = EFAULT;
	  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	  return -1;
	}

      bzero (buf, sizeof (*buf));
      //buf->f_fsid.val[1] = 0;
      buf->f_fsid.val[0] = (long) handler;
      dv = (APTR) LockDosList(LDF_DEVICES | LDF_READ);
      while ((dv = (APTR) NextDosEntry((APTR) dv, LDF_DEVICES | LDF_READ)))
	{
	  if ((struct MsgPort *) dv->dvi_Task == handler)
	    {
	      unsigned char *name = BTOCPTR (dv->dvi_Name);

	      buf->f_mntonname[0] = '/';
	      bcopy (name + 1, buf->f_mntonname + 1, name[0]);
	      //buf->f_mntonname[name[0] + 1] = 0;
	      break;
	    }
	}
      UnLockDosList(LDF_DEVICES | LDF_READ);

      if (! dv)
	{
	  errno = EOPNOTSUPP;
	  KPRINTF (("&errno = %lx, errno = %ld\n", &errno, errno));
	  return -1;
	}

      internal_statfs64 (buf);

      return 0;
    }
}
