/*
 * Copyright (c) 1989 The Regents of the University of California.
 * Copyright (C) 2021 Harry Sintonen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by the University of California, Berkeley and its contributors''
 * in the documentation or other materials provided with the distribution
 * and in all advertising materials mentioning features or use of this
 * software. Neither the name of the University nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)strftime.c  5.8 (Berkeley) 6/1/90";
#endif /* LIBC_SCCS and not lint */

/*
 * NOTE: I used the older version of this file, since the 4.3bsd-net2
 *       version uses mktime() from ctime.c, and ctime.c is badly suited
 *       for inclusion in a shared library (too much static stuff...)
 */

#define _KERNEL
#include "ixemul.h"

#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <tzfile.h>
#include <string.h>
#ifdef USE_WIDECHAR
#include <wchar.h>
#endif

#ifndef USE_WIDECHAR
#define CHARTYPE char
#define MKS(x) x
#define FUNCNAME strftime
#else
#define CHARTYPE wchar_t
#define MKS(x) L ## x
#define FUNCNAME wcsftime
#endif

static const CHARTYPE *afmt[] = {
	MKS("Sun"), MKS("Mon"), MKS("Tue"), MKS("Wed"), MKS("Thu"), MKS("Fri"), MKS("Sat"),
};
static const CHARTYPE *Afmt[] = {
	MKS("Sunday"), MKS("Monday"), MKS("Tuesday"), MKS("Wednesday"), MKS("Thursday"), MKS("Friday"),
	MKS("Saturday"),
};
static const CHARTYPE *bfmt[] = {
	MKS("Jan"), MKS("Feb"), MKS("Mar"), MKS("Apr"), MKS("May"), MKS("Jun"), MKS("Jul"), MKS("Aug"), MKS("Sep"),
	MKS("Oct"), MKS("Nov"), MKS("Dec"),
};
static const CHARTYPE *Bfmt[] = {
	MKS("January"), MKS("February"), MKS("March"), MKS("April"), MKS("May"), MKS("June"), MKS("July"),
	MKS("August"), MKS("September"), MKS("October"), MKS("November"), MKS("December"),
};

struct localvars
{
	CHARTYPE *u_pt;
	size_t    u_gsize;
};

static size_t _fmt(const CHARTYPE *, const struct tm *, struct localvars *);
static int _conv (int, int, CHARTYPE, struct localvars *);
static int _add (const CHARTYPE *, struct localvars *);

size_t FUNCNAME(CHARTYPE *s, size_t maxsize, const CHARTYPE *format, const struct tm *t)
{
	size_t res = 0;
	struct localvars lv = {s, maxsize};

	if (maxsize < 1)
		return 0;
	if (_fmt(format, t, &lv)) {
		*lv.u_pt = MKS('\0');
		res = maxsize - lv.u_gsize;
	}
	return res;
}

static size_t _fmt(const CHARTYPE *format, const struct tm *t, struct localvars *lv)
{
	for (; *format; ++format) {
		if (*format == MKS('%'))
			switch (*++format) {
			case MKS('\0'):
				--format;
				break;
			case MKS('A'):
				if (t->tm_wday < 0 || t->tm_wday > 6)
					return(0);
				if (!_add(Afmt[t->tm_wday], lv))
					return(0);
				continue;
			case MKS('a'):
				if (t->tm_wday < 0 || t->tm_wday > 6)
					return(0);
				if (!_add(afmt[t->tm_wday], lv))
					return(0);
				continue;
			case MKS('B'):
				if (t->tm_mon < 0 || t->tm_mon > 11)
					return(0);
				if (!_add(Bfmt[t->tm_mon], lv))
					return(0);
				continue;
			case MKS('b'):
			case MKS('h'):
				if (t->tm_mon < 0 || t->tm_mon > 11)
					return(0);
				if (!_add(bfmt[t->tm_mon], lv))
					return(0);
				continue;
			case MKS('C'):
				if (!_fmt(MKS("%a %b %e %H:%M:%S %Y"), t, lv))
					return(0);
				continue;
			case MKS('c'):
				if (!_fmt(MKS("%m/%d/%y %H:%M:%S"), t, lv))
					return(0);
				continue;
			case MKS('e'):
				if (!_conv(t->tm_mday, 2, MKS(' '), lv))
					return(0);
				continue;
			case MKS('F'): // Equivalent to %Y-%m-%d (the ISO 8601 date format). (C99)
				if (!_fmt(MKS("%Y-%m-%d"), t, lv))
					return(0);
				continue;
			case MKS('D'):
				if (!_fmt(MKS("%m/%d/%y"), t, lv))
					return(0);
				continue;
			case MKS('d'):
				if (!_conv(t->tm_mday, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('H'):
				if (!_conv(t->tm_hour, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('I'):
				if (!_conv(t->tm_hour % 12 ?
				    t->tm_hour % 12 : 12, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('j'):
				if (!_conv(t->tm_yday + 1, 3, MKS('0'), lv))
					return(0);
				continue;
			case MKS('k'):
				if (!_conv(t->tm_hour, 2, MKS(' '), lv))
					return(0);
				continue;
			case MKS('l'):
				if (!_conv(t->tm_hour % 12 ?
				    t->tm_hour % 12 : 12, 2, MKS(' '), lv))
					return(0);
				continue;
			case MKS('M'):
				if (!_conv(t->tm_min, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('m'):
				if (!_conv(t->tm_mon + 1, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('n'):
				if (!_add(MKS("\n"), lv))
					return(0);
				continue;
			case MKS('p'):
				if (!_add(t->tm_hour >= 12 ? MKS("PM") : MKS("AM"), lv))
					return(0);
				continue;
			case MKS('R'):
				if (!_fmt(MKS("%H:%M"), t, lv))
					return(0);
				continue;
			case MKS('r'):
				if (!_fmt(MKS("%I:%M:%S %p"), t, lv))
					return(0);
				continue;
			case MKS('S'):
				if (!_conv(t->tm_sec, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('T'):
			case MKS('X'):
				if (!_fmt(MKS("%H:%M:%S"), t, lv))
					return(0);
				continue;
			case MKS('t'):
				if (!_add(MKS("\t"), lv))
					return(0);
				continue;
			case MKS('U'):
				if (!_conv((t->tm_yday + 7 - t->tm_wday) / 7,
				    2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('u'): // Weekday as a decimal number, where Monday is 1 (ISO 8601 format) (range [1-7])
				if (!_conv(1 + (t->tm_wday + 6) % 7,
				    1, MKS('0'), lv))
					return(0);
				continue;
			case MKS('W'):
				if (!_conv((t->tm_yday + 7 -
				    (t->tm_wday ? (t->tm_wday - 1) : 6))
				    / 7, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('w'):
				if (!_conv(t->tm_wday, 1, MKS('0'), lv))
					return(0);
				continue;
			case MKS('x'):
				if (!_fmt(MKS("%m/%d/%y"), t, lv))
					return(0);
				continue;
			case MKS('y'):
				if (!_conv((t->tm_year + TM_YEAR_BASE)
				    % 100, 2, MKS('0'), lv))
					return(0);
				continue;
			case MKS('Y'):
				if (!_conv(t->tm_year + TM_YEAR_BASE, 4, MKS('0'), lv))
					return(0);
				continue;
			case MKS('z'): // The +hhmm or -hhmm numeric timezone (that is, the hour and minute offset from UTC). (SU)
				{
					const unsigned int mins = abs(t->tm_gmtoff) / 60;
					const unsigned int hours = mins / 60;
					if (!_add(t->tm_gmtoff < 0 ? MKS("-") : MKS("+"), lv) ||
					    !_conv(hours % 24, 2, MKS('0'), lv) ||
					    !_conv(mins % 60, 2, MKS('0'), lv))
						return(0);
				}
				continue;
			case MKS('Z'):
#ifndef USE_WIDECHAR
				if (!t->tm_zone || !_add(t->tm_zone, lv))
					return(0);
#else
				if (!t->tm_zone)
					return(0);
				{
					const char *src = t->tm_zone;
					size_t ct = mbsrtowcs(lv->u_pt, &src, lv->u_gsize, NULL);
					if (ct == (size_t) -1)
						return(0);
					lv->u_gsize -= ct;
					lv->u_pt += ct;
				}
#endif
				continue;
			case MKS('%'):
			/*
			 * X311J/88-090 (4.12.3.5): if conversion char is
			 * undefined, behavior is undefined.  Print out the
			 * character itself as printf(3) does.
			 */
			default:
				break;
		}
		if (!lv->u_gsize--)
			return(0);
		*lv->u_pt++ = *format;
	}
	return(lv->u_gsize);
}

static int _conv(int n, int digits, CHARTYPE pad, struct localvars *lv)
{
	CHARTYPE buf[10];
	register CHARTYPE *p;

	buf[sizeof(buf)/sizeof(buf[0]) - 1] = MKS('\0');
	for (p = buf + sizeof(buf)/sizeof(buf[0]) - 2; n > 0 && p > buf; n /= 10, --digits)
		*p-- = n % 10 + MKS('0');
	while (p > buf && digits-- > 0)
		*p-- = pad;
	return(_add(++p, lv));
}

static int _add(const CHARTYPE *str, struct localvars *lv)
{
	for (;; ++lv->u_pt, --lv->u_gsize) {
		if (!lv->u_gsize)
			return(0);
		if (!(*lv->u_pt = *str++))
			return(1);
	}
}
